# This file contains alternative implementations of some PyDRex functions
# Line 14: Original strain, to be used with the original pathline
# Line 120: strain modified to accommodate for the new pathline
# Line 240: strain using Scipy solve_ivp, requiring the new pathline
# Line 327: Original eigen, necessary for the original pipar
# Line 365: Original pipar
# Line 415: New pipar, mainly a re-write of isacalc
# Line 473: Original pathline
# Line 530: New pathline, using Scipy solve_ivp
# Line 577: DRex matching the original implementation of pathline and strain
# Line 609: DRex matching the new implementation of pathline and strain
# Original strain, to be used with the original pathline
def strain(step, stream_Lij, stream_e, stream_dt, stream_alpha, dictGlobals=None):
    """Uncomment this block to use Ray
    if dictGlobals:
        acs0 = dictGlobals['acs0']
        chi = dictGlobals['chi']
        size = dictGlobals['size']
    """
    fse = np.identity(3)
    odf = 1 / size * np.ones(size)
    odf_ens = 1 / size * np.ones(size)
    acs = np.copy(acs0)
    acs_ens = np.copy(acs0)
    # step-number of steps used to construct the streamline
    for currStep in range(step, -1, -1):
        # Lij at time from record
        L = np.ascontiguousarray(stream_Lij[:, :, currStep])
        t_loc = stream_dt[currStep]
        epsnot = stream_e[currStep]  # epsnot: reference strain rate
        alpha = stream_alpha[currStep]  # alpha at current location
        e = (L + np.transpose(L)) / 2
        # time stepping for LPO calculation
        dt = min(t_loc, 1e-2 / epsnot)
        # number of iterations in the LPO loop
        N_strain = int(round(t_loc / dt))
        # LPO loop on the point on the pathline
        for nn in range(N_strain):
            # CALL 1/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acs, acs_ens, fse, odf, odf_ens, alpha
            )
            kfse1 = np.dot(L, fse) * dt
            kodf1 = dotodf * dt * epsnot
            kodf1_ens = dotodf_ens * dt * epsnot
            kac1 = dotacs * dt * epsnot
            kac1_ens = dotacs_ens * dt * epsnot
            fsei = fse + 0.5 * kfse1
            acsi = np.clip(acs + 0.5 * kac1, -1, 1)
            acsi_ens = np.clip(acs_ens + 0.5 * kac1_ens, -1, 1)
            odfi = np.clip(odf + 0.5 * kodf1, 0, None)
            odfi /= np.sum(odfi)
            odfi_ens = np.clip(odf_ens + 0.5 * kodf1_ens, 0, None)
            odfi_ens /= np.sum(odfi_ens)
            # CALL 2/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse2 = np.dot(L, fsei) * dt
            kodf2 = dotodf * dt * epsnot
            kodf2_ens = dotodf_ens * dt * epsnot
            kac2 = dotacs * dt * epsnot
            kac2_ens = dotacs_ens * dt * epsnot
            fsei = fse + 0.5 * kfse2
            acsi = np.clip(acs + 0.5 * kac2, -1, 1)
            acsi_ens = np.clip(acs_ens + 0.5 * kac2_ens, -1, 1)
            odfi = np.clip(odf + 0.5 * kodf2, 0, None)
            odfi /= np.sum(odfi)
            odfi_ens = np.clip(odf_ens + 0.5 * kodf2_ens, 0, None)
            odfi_ens /= np.sum(odfi_ens)
            # CALL 3/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse3 = np.dot(L, fsei) * dt
            kodf3 = dotodf * dt * epsnot
            kodf3_ens = dotodf_ens * dt * epsnot
            kac3 = dotacs * dt * epsnot
            kac3_ens = dotacs_ens * dt * epsnot
            fsei = fse + kfse3
            acsi = np.clip(acs + kac3, -1, 1)
            acsi_ens = np.clip(acs_ens + kac3_ens, -1, 1)
            odfi = np.clip(odf + kodf3, 0, None)
            odfi /= np.sum(odfi)
            odfi_ens = np.clip(odf_ens + kodf3_ens, 0, None)
            odfi_ens /= np.sum(odfi_ens)
            # CALL 4/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse4 = np.dot(L, fsei) * dt
            kodf4 = dotodf * dt * epsnot
            kodf4_ens = dotodf_ens * dt * epsnot
            kac4 = dotacs * dt * epsnot
            kac4_ens = dotacs_ens * dt * epsnot
            fse += (kfse1 / 2 + kfse2 + kfse3 + kfse4 / 2) / 3
            acs = np.clip(acs + (kac1 / 2 + kac2 + kac3 + kac4 / 2) / 3, -1, 1)
            acs_ens = np.clip(
                acs_ens + (kac1_ens / 2 + kac2_ens + kac3_ens + kac4_ens / 2) / 3, -1, 1
            )
            odf += (kodf1 / 2 + kodf2 + kodf3 + kodf4 / 2) / 3
            odf_ens += (kodf1_ens / 2 + kodf2_ens + kodf3_ens + kodf4_ens / 2) / 3
            odf /= np.sum(odf)
            odf_ens /= np.sum(odf_ens)
            # Grain-boundary sliding
            mask = odf <= chi / size
            mask_ens = odf_ens <= chi / size
            acs[mask, :, :] = acs0[mask, :, :]
            acs_ens[mask_ens, :, :] = acs0[mask_ens, :, :]
            odf[mask] = chi / size
            odf_ens[mask_ens] = chi / size
            odf /= np.sum(odf)
            odf_ens /= np.sum(odf_ens)
    return fse, acs, acs_ens, odf, odf_ens


# strain modified to accommodate for the new pathline
def strain(pathTime, pathDense, dictGlobals=None):
    """Uncomment this block to use Ray
    if dictGlobals:
        acs0 = dictGlobals['acs0']
        chi = dictGlobals['chi']
        gridCoords = dictGlobals['gridCoords']
        iDefMech = dictGlobals['iDefMech']
        size = dictGlobals['size']
    """
    fse = np.identity(3)
    acs = acs0.copy()
    acs_ens = acs0.copy()
    odf = np.ones(size) / size
    odf_ens = np.ones(size) / size
    for time in reversed(pathTime):
        if isInside(pathDense(time), dictGlobals):
            currTime = time
            break
    while currTime < pathTime[0]:
        currPoint = pathDense(currTime)
        currVel = interpVel(currPoint, dictGlobals)
        L, epsnot = interpVelGrad(currPoint, dictGlobals)
        e = (L + L.transpose()) / 2
        alpha = iDefMech(*currPoint)
        indLeft = [
            np.searchsorted(gridCoords[x], currPoint[x]) for x in range(currPoint.size)
        ]
        gridStep = []
        for coord, ind in zip(gridCoords, indLeft):
            try:
                gridStep.append(coord[ind + 1] - coord[ind])
            except IndexError:
                gridStep.append(coord[ind] - coord[ind - 1])
        dtPathline = min(gridStep) / 4 / norm(currVel, ord=2)
        dtPathline = min(dtPathline, pathTime[0] - currTime)
        # time stepping for LPO calculation
        dt = min(dtPathline, 1e-2 / epsnot)
        # number of iterations in the LPO loop
        nbIter = int(dtPathline / dt)
        # LPO loop on the point on the pathline
        for iter in range(nbIter):
            # CALL 1/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acs, acs_ens, fse, odf, odf_ens, alpha
            )
            kfse1 = np.dot(L, fse) * dt
            kodf1 = dotodf * dt * epsnot
            kodf1_ens = dotodf_ens * dt * epsnot
            kac1 = dotacs * dt * epsnot
            kac1_ens = dotacs_ens * dt * epsnot
            fsei = fse + 0.5 * kfse1
            acsi = np.clip(acs + 0.5 * kac1, -1, 1)
            acsi_ens = np.clip(acs_ens + 0.5 * kac1_ens, -1, 1)
            odfi = np.clip(odf + 0.5 * kodf1, 0, None)
            odfi /= odfi.sum()
            odfi_ens = np.clip(odf_ens + 0.5 * kodf1_ens, 0, None)
            odfi_ens /= odfi_ens.sum()
            # CALL 2/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse2 = np.dot(L, fsei) * dt
            kodf2 = dotodf * dt * epsnot
            kodf2_ens = dotodf_ens * dt * epsnot
            kac2 = dotacs * dt * epsnot
            kac2_ens = dotacs_ens * dt * epsnot
            fsei = fse + 0.5 * kfse2
            acsi = np.clip(acs + 0.5 * kac2, -1, 1)
            acsi_ens = np.clip(acs_ens + 0.5 * kac2_ens, -1, 1)
            odfi = np.clip(odf + 0.5 * kodf2, 0, None)
            odfi /= odfi.sum()
            odfi_ens = np.clip(odf_ens + 0.5 * kodf2_ens, 0, None)
            odfi_ens /= odfi_ens.sum()
            # CALL 3/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse3 = np.dot(L, fsei) * dt
            kodf3 = dotodf * dt * epsnot
            kodf3_ens = dotodf_ens * dt * epsnot
            kac3 = dotacs * dt * epsnot
            kac3_ens = dotacs_ens * dt * epsnot
            fsei = fse + kfse3
            acsi = np.clip(acs + kac3, -1, 1)
            acsi_ens = np.clip(acs_ens + kac3_ens, -1, 1)
            odfi = np.clip(odf + kodf3, 0, None)
            odfi /= odfi.sum()
            odfi_ens = np.clip(odf_ens + kodf3_ens, 0, None)
            odfi_ens /= odfi_ens.sum()
            # CALL 4/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse4 = np.dot(L, fsei) * dt
            kodf4 = dotodf * dt * epsnot
            kodf4_ens = dotodf_ens * dt * epsnot
            kac4 = dotacs * dt * epsnot
            kac4_ens = dotacs_ens * dt * epsnot
            fse += (kfse1 / 2 + kfse2 + kfse3 + kfse4 / 2) / 3
            acs = np.clip(acs + (kac1 / 2 + kac2 + kac3 + kac4 / 2) / 3, -1, 1)
            acs_ens = np.clip(
                acs_ens + (kac1_ens / 2 + kac2_ens + kac3_ens + kac4_ens / 2) / 3, -1, 1
            )
            odf += (kodf1 / 2 + kodf2 + kodf3 + kodf4 / 2) / 3
            odf_ens += (kodf1_ens / 2 + kodf2_ens + kodf3_ens + kodf4_ens / 2) / 3
            odf /= odf.sum()
            odf_ens /= odf_ens.sum()
            # Grain-boundary sliding
            mask = odf < chi / size
            mask_ens = odf_ens < chi / size
            acs[mask, :, :] = acs0[mask, :, :]
            acs_ens[mask_ens, :, :] = acs0[mask_ens, :, :]
            odf[mask] = chi / size
            odf_ens[mask_ens] = chi / size
            odf /= odf.sum()
            odf_ens /= odf_ens.sum()
        currTime += nbIter * dt
    return fse, acs, acs_ens, odf, odf_ens


# strain using Scipy solve_ivp, requiring the new pathline
def strain(pathTime, pathDense, dictGlobals=None):
    """Uncomment this block to use Ray
    if dictGlobals:
        acs0 = dictGlobals['acs0']
        chi = dictGlobals['chi']
        gridCoords = dictGlobals['gridCoords']
        iDefMech = dictGlobals['iDefMech']
        size = dictGlobals['size']
    """

    def extractVars(y):
        fse = y[:9].copy().reshape(3, 3)
        acs = y[9 : size * 9 + 9].copy().reshape(size, 3, 3).clip(-1, 1)
        acs_ens = y[size * 9 + 9 : size * 18 + 9].copy().reshape(size, 3, 3)
        acs_ens.clip(-1, 1, out=acs_ens)
        odf = y[size * 18 + 9 : size * 19 + 9].copy().clip(0, None)
        odf /= odf.sum()
        odf_ens = y[size * 19 + 9 : size * 20 + 9].copy().clip(0, None)
        odf_ens /= odf_ens.sum()
        return fse, acs, acs_ens, odf, odf_ens

    def grainBoundarySliding(acs, acs_ens, odf, odf_ens):
        mask = odf < chi / size
        mask_ens = odf_ens < chi / size
        acs[mask, :, :] = acs0[mask, :, :]
        acs_ens[mask_ens, :, :] = acs0[mask_ens, :, :]
        odf[mask] = chi / size
        odf /= odf.sum()
        odf_ens[mask_ens] = chi / size
        odf_ens /= odf_ens.sum()
        return acs, acs_ens, odf, odf_ens

    def derivIVP(t, y):
        fsei, acsi, acsi_ens, odfi, odfi_ens = extractVars(y)
        dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
            L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
        )
        return np.hstack(
            (
                np.dot(L, fsei).flatten(),
                dotacs.flatten() * epsnot,
                dotacs_ens.flatten() * epsnot,
                dotodf * epsnot,
                dotodf_ens * epsnot,
            )
        )

    def eventIVP(t, y):
        nonlocal alpha, e, epsnot, fse, L
        fse, acs, acs_ens, odf, odf_ens = extractVars(y)
        acs, acs_ens, odf, odf_ens = grainBoundarySliding(acs, acs_ens, odf, odf_ens)
        y[9:] = np.hstack((acs.flatten(), acs_ens.flatten(), odf, odf_ens))
        currPoint = pathDense(t)
        L, epsnot = interpVelGrad(currPoint, dictGlobals)
        e = (L + L.transpose()) / 2
        alpha = iDefMech(*currPoint)
        return -1

    fse = np.identity(3)
    for time in reversed(pathTime):
        if isInside(pathDense(time), dictGlobals):
            currTime = time
            break
    currPoint = pathDense(currTime)
    L, epsnot = interpVelGrad(currPoint, dictGlobals)
    e = (L + L.transpose()) / 2
    alpha = iDefMech(*currPoint)

    currVel = interpVel(currPoint, dictGlobals)
    indLeft = [
        np.searchsorted(gridCoords[x], currPoint[x]) for x in range(currPoint.size)
    ]
    gridStep = []
    for coord, ind in zip(gridCoords, indLeft):
        try:
            gridStep.append(coord[ind + 1] - coord[ind])
        except IndexError:
            gridStep.append(coord[ind] - coord[ind - 1])
    dtPathline = min(gridStep) / 4 / norm(currVel, ord=2)
    dt = min(dtPathline, pathTime[0] - currTime, 1e-2 / epsnot)

    sol = solve_ivp(
        derivIVP,
        [currTime, pathTime[0]],
        np.hstack(
            (
                fse.flatten(),
                acs0.copy().flatten(),
                acs0.copy().flatten(),
                np.ones(size) / size,
                np.ones(size) / size,
            )
        ),
        method="RK45",
        first_step=dt,
        max_step=5e12,
        t_eval=[pathTime[0]],
        events=[eventIVP],
        atol=1e-8,
        rtol=1e-5,
    )
    fse, acs, acs_ens, odf, odf_ens = extractVars(sol.y.squeeze())
    acs, acs_ens, odf, odf_ens = grainBoundarySliding(acs, acs_ens, odf, odf_ens)
    return fse, acs, acs_ens, odf, odf_ens


# Original eigen
# Find 3 eigenvalues of velocity gradient tensor
def eigen(L):
    Id = np.identity(3)
    Q = 0
    for ii in range(2):
        for jj in range(ii + 1, 3):
            Q -= (L[ii, ii] * L[jj, jj] - L[ii, jj] * L[jj, ii]) / 3
    R = np.linalg.det(L) / 2
    if abs(Q) < 1e-9:
        F = -np.ones((3, 3))
    elif Q**3 - R**2 >= 0:
        theta = np.arccos(R / Q**1.5)
        L1 = -2 * np.sqrt(Q) * np.cos(theta / 3)
        L2 = -2 * np.sqrt(Q) * np.cos((theta + 2 * np.pi) / 3)
        L3 = -2 * np.sqrt(Q) * np.cos((theta + 4 * np.pi) / 3)
        if abs(L1 - L2) < 1e-13:
            L1 = L2
        if abs(L3 - L2) < 1e-13:
            L3 = L2
        if abs(L1 - L3) < 1e-13:
            L1 = L3
        if L1 > L2 and (L1 > L3 or L2 == L3):
            F = np.dot(L - L2 * Id, L - L3 * Id)
        elif L2 > L3 and (L2 > L1 or L3 == L1):
            F = np.dot(L - L3 * Id, L - L1 * Id)
        elif L3 > L1 and (L3 > L2 or L1 == L2):
            F = np.dot(L - L1 * Id, L - L2 * Id)
        else:
            F = np.zeros((3, 3))
    else:
        xx = (np.sqrt(R**2 - Q**3) + abs(R)) ** (1 / 3)
        L1 = -(xx + Q / xx) if R == 0 else -np.sign(R) * (xx + Q / xx)
        L2 = L3 = -L1 / 2
        F = np.dot(L - L2 * Id, L - L3 * Id) if L1 > 1e-9 else np.zeros((3, 3))
    return F


# Original pipar
def pipar(currPoint, dictGlobals=None):
    # isacalc: Calculates ISA Orientation at grid point
    def isacalc(L, veloc):
        nonlocal GOL
        F = eigen(L)
        print(eigh(np.dot(F, np.transpose(F))), eigh(Ua))
        if np.sum(F) == -9:
            isa = -np.ones(3)
        elif np.sum(np.absolute(F)) == 0:
            isa = np.zeros(3)
            GOL = -1
        else:
            isa = eigh(np.dot(F, np.transpose(F)))[1][:, -1]
        return isa if dim == 3 else isa[::2]

    GOL = 10
    # local velocity vector and orientation of infinite strain axis
    veloc = interpVel(currPoint, dictGlobals)
    dt = 1e-2 / norm(veloc, ord=2)
    # previous point on the streamline
    prevPoint = currPoint - dt * veloc
    if isInside(prevPoint, dictGlobals) is False:
        return -10
    veloc = interpVel(prevPoint, dictGlobals)
    # epsnot: reference strain rate
    L, epsnot = interpVelGrad(prevPoint, dictGlobals)
    L /= epsnot
    # calculation of the ISA
    isa = veloc / norm(veloc, ord=2) if np.sum(isacalc(L)) == -3 else isacalc(L)
    # angle between ISA and flow direction
    thetaISA = np.arccos(np.sum(veloc / norm(veloc, ord=2) * isa))
    # next point on the streamline
    nextPoint = currPoint + dt * veloc
    if isInside(nextPoint, dictGlobals) is False:
        return -10
    veloc = veloc / norm(veloc, ord=2)
    L, epsnot = interpVelGrad(nextPoint, dictGlobals)
    L /= epsnot
    # calculation of the ISA
    isa = veloc if np.sum(isacalc(L)) == -3 else isacalc(L)
    L, epsnot = interpVelGrad(currPoint, dictGlobals)
    # angle between ISA and flow direction
    thetaISA = abs(thetaISA - np.arccos(np.sum(veloc / norm(veloc, ord=2) * isa)))
    if thetaISA > np.pi:
        thetaISA -= np.pi
    return min(GOL, thetaISA / 2 / dt / epsnot)


# New pipar, mainly a re-write of isacalc
def pipar(currPoint, dictGlobals=None):
    # Calculates ISA Orientation at grid point
    def isacalc(L, veloc):
        nonlocal GOL
        # Kaminski, Ribe & Browaeys (2004): Appendix B
        evals = eigvalsh(L)
        if np.sum(np.absolute(evals) < 1e-9) >= 2 - veloc.size % 2:
            assert (
                abs(evals[0] * evals[1] + evals[0] * evals[2] + evals[1] * evals[2])
                < 1e-9
            )
            return veloc / norm(veloc, ord=2)
        else:
            ind = np.argsort(np.absolute(evals))[-1]
            if np.isreal(evals[ind]):
                a = (ind + 1) % 3
                b = (ind + 2) % 3
                Id = np.identity(3)
                Fa = (
                    np.dot(L - evals[a] * Id, L - evals[b] * Id)
                    / (evals[ind] - evals[a])
                    / (evals[ind] - evals[b])
                )
                Ua = np.dot(np.transpose(Fa), Fa)
                return eigh(Ua)[1][:: 4 - veloc.size, -1]
            else:
                GOL = -1
                return np.zeros(veloc.size)

    GOL = 10
    # local velocity vector and orientation of infinite strain axis
    veloc = interpVel(currPoint, dictGlobals)
    dt = 1e-2 / norm(veloc, ord=2)
    # previous point on the streamline
    prevPoint = currPoint - dt * veloc
    if isInside(prevPoint, dictGlobals) is False:
        return -10
    veloc = interpVel(prevPoint, dictGlobals)
    # epsnot: reference strain rate
    L, epsnot = interpVelGrad(prevPoint, dictGlobals)
    L /= epsnot
    # calculation of the ISA
    isa = isacalc(L, veloc)
    # angle between ISA and flow direction
    thetaISA = np.arccos(np.sum(veloc / norm(veloc, ord=2) * isa))
    # next point on the streamline
    nextPoint = currPoint + dt * veloc
    if isInside(nextPoint, dictGlobals) is False:
        return -10
    veloc = interpVel(nextPoint, dictGlobals)
    L, epsnot = interpVelGrad(nextPoint, dictGlobals)
    L /= epsnot
    # calculation of the ISA
    isa = isacalc(L, veloc)
    L, epsnot = interpVelGrad(currPoint, dictGlobals)
    # angle between ISA and flow direction
    thetaISA = abs(thetaISA - np.arccos(np.sum(veloc / norm(veloc, ord=2) * isa)))
    if thetaISA > np.pi:
        thetaISA -= np.pi
    return min(GOL, thetaISA / 2 / dt / epsnot)


# Original pathline
def pathline(currPoint, dictGlobals=None):
    """Uncomment this block to use Ray
    if dictGlobals:
        gridCoords = dictGlobals['gridCoords']
        gridNodes = dictGlobals['gridNodes']
        iDefMech = dictGlobals['iDefMech']
    """
    stream_Lij = np.zeros((3, 3, 10000))
    stream_dt = np.zeros(10000)
    stream_e = np.zeros(10000)
    stream_alpha = np.zeros(10000)
    max_strain = 0
    step = -1
    # Construction of the streamline
    while max_strain < 10 and step < 1e4:
        step += 1
        currVel = interpVel(currPoint, dictGlobals)
        L, epsnot = interpVelGrad(currPoint, dictGlobals)
        dt = (
            min(
                abs(gridCoords[x][currInd[x] + 1] - gridCoords[x][currInd[x]])
                for x in range(len(currInd))
            )
            / 4
            / norm(currVel, ord=2)
        )
        # Record of the local velocity gradient tensor and time spent at that
        # point
        stream_Lij[:, :, step] = L
        stream_dt[step] = dt
        stream_e[step] = epsnot
        stream_alpha[step] = iDefMech(*currPoint)
        max_strain += dt * epsnot
        k1 = -currVel * dt
        newPoint = currPoint + 0.5 * k1
        if isInside(newPoint, dictGlobals) is False:
            break
        currVel = interpVel(newPoint, dictGlobals)
        k2 = -currVel * dt
        newPoint = currPoint + 0.5 * k2
        if isInside(newPoint, dictGlobals) is False:
            break
        currVel = interpVel(newPoint, dictGlobals)
        k3 = -currVel * dt
        newPoint = currPoint + k3
        if isInside(newPoint, dictGlobals) is False:
            break
        currVel = interpVel(newPoint, dictGlobals)
        currPoint += (k1 / 2 + k2 + k3 - currVel * dt / 2) / 3
        if isInside(currPoint, dictGlobals) is False:
            break
        for i in range(len(currPoint)):
            while gridCoords[i][currInd[i]] > currPoint[i] and currInd[i] > 0:
                currInd[i] -= 1
            while (
                gridCoords[i][currInd[i] + 1] < currPoint[i]
                and currInd[i] < gridNodes[i] - 2
            ):
                currInd[i] += 1
    return (
        step,
        stream_Lij[:, :, : step + 1],
        stream_e[: step + 1],
        stream_dt[: step + 1],
        stream_alpha[: step + 1],
    )


# New pathline, using Scipy solve_ivp
def pathline(currPoint, dictGlobals=None):
    def ivpFunc(time, pointCoords, dictGlobals=None):
        if isInside(pointCoords, dictGlobals):
            return interpVel(pointCoords, dictGlobals)
        else:
            return np.zeros(pointCoords.shape)

    def ivpJac(time, pointCoords, dictGlobals=None):
        if isInside(pointCoords, dictGlobals):
            L = interpVelGrad(pointCoords, dictGlobals, True)
            if pointCoords.size == 2:
                return np.array([[L[0, 0], L[0, 2]], [L[2, 0], L[2, 2]]])
            else:
                return L
        else:
            return np.zeros((pointCoords.size, pointCoords.size))

    def maxStrain(time, pointCoords, dictGlobals=None):
        nonlocal eventTime, eventTimePrev, eventStrain, eventStrainPrev, eventFlag
        if eventFlag:
            return (eventStrain if time == eventTime else eventStrainPrev) - 10
        elif isInside(pointCoords, dictGlobals):
            L, epsnot = interpVelGrad(pointCoords, dictGlobals)
            eventStrainPrev = eventStrain
            eventStrain += abs(time - eventTime) * epsnot
            if eventStrain >= 10:
                eventFlag = True
            eventTimePrev = eventTime
            eventTime = time
            return eventStrain - 10
        else:
            return 0

    maxStrain.terminal = True
    eventStrain = eventTime = 0
    eventStrainPrev = eventTimePrev = None
    eventFlag = False
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        sol = solve_ivp(
            ivpFunc,
            [0, -100e6 * 365.25 * 8.64e4],
            currPoint,
            method="RK45",
            first_step=1e11,
            max_step=np.inf,
            t_eval=None,
            events=[maxStrain],
            args=(dictGlobals,),
            dense_output=True,
            jac=ivpJac,
            atol=1e-8,
            rtol=1e-5,
        )
    return sol.t, sol.sol


# DRex matching the original implementation of pathline and strain
def DRex(locInd, dictGlobals=None):
    """Uncomment this block to use Ray
    if dictGlobals:
        gridCoords = dictGlobals['gridCoords']
    """
    # Initial location of the grid point
    currCoords = np.array(
        [(coord[ind] + coord[ind + 1]) / 2 for coord, ind in zip(gridCoords, locInd)]
    )
    # Grain Orientation Lag
    GOL = pipar(currCoords, dictGlobals)
    # Backward calculation of the pathline for each tracer
    step, stream_Lij, stream_e, stream_dt, stream_alpha = pathline(
        list(locInd), currCoords, dictGlobals
    )
    # Inward calculation of the LPO
    # Random initial LPO
    Fij, odf, odf_ens, acs, acs_ens = strain(
        step, stream_Lij, stream_e, stream_dt, stream_alpha, dictGlobals
    )
    # Left-stretch tensor for FSE calculation
    LSij = np.dot(Fij, np.transpose(Fij))
    eigval, eigvects = eigh(LSij)
    # pick up the orientation of the long axis of the FSE
    phi_fse = np.arctan2(eigvects[-1, -1], eigvects[0, -1])
    # natural strain = ln(a / c) where a is the long axis = max(eigval) ** 0.5
    ln_fse = np.log(eigval[-1] / eigval[0]) / 2
    # Cijkl tensor (using Voigt average)
    Sav = voigt(odf, odf_ens, acs, acs_ens)
    # percentage anisotropy and orientation of hexagonal symmetry axis
    perc_a, phi_a, radani, percani = decsym(Sav)
    return locInd[::-1], radani, percani, GOL, ln_fse, phi_fse, phi_a, perc_a


# DRex matching the new implementation of pathline and strain
def DRex(locInd, dictGlobals=None):
    """Uncomment this block to use Ray
    if dictGlobals:
        gridCoords = dictGlobals['gridCoords']
    """
    # Initial location of the grid point
    currCoords = np.array(
        [(coord[ind] + coord[ind + 1]) / 2 for coord, ind in zip(gridCoords, locInd)]
    )
    # Grain Orientation Lag
    GOL = pipar(currCoords, dictGlobals)
    # Backward calculation of the pathline for each tracer
    pathTime, pathDense = pathline(currCoords, dictGlobals)
    # Inward calculation of the LPO | Random initial LPO
    Fij, odf, odf_ens, acs, acs_ens = strain(pathTime, pathDense, dictGlobals)
    # Left-stretch tensor for FSE calculation
    LSij = np.dot(Fij, np.transpose(Fij))
    eigval, eigvects = eigh(LSij)
    # pick up the orientation of the long axis of the FSE
    phi_fse = np.arctan2(eigvects[-1, -1], eigvects[0, -1])
    # natural strain = ln(a / c) where a is the long axis = max(eigval) ** 0.5
    ln_fse = np.log(eigval[-1] / eigval[0]) / 2
    # Cijkl tensor (using Voigt average)
    Sav = voigt(odf, odf_ens, acs, acs_ens)
    # percentage anisotropy and orientation of hexagonal symmetry axis
    perc_a, phi_a, radani, percani = decsym(Sav)
    return locInd[::-1], radani, percani, GOL, ln_fse, phi_fse, phi_a, perc_a



"""> PyDRex: tests for configuration file format."""

import pathlib as pl

import numpy as np

from pydrex import core as _core
from pydrex import io as _io
from pydrex import velocity as _velocity


def test_specfile():
    """Test TOML spec file parsing."""
    config = _io.parse_config(_io.data("specs") / "spec.toml")
    assert list(config.keys()) == ["name", "input", "output", "parameters"]
    _input = config["input"]
    # TODO: Add some example mesh/path files and use them in another test.
    assert _input["mesh"] is None
    assert _input["locations_final"] is None
    assert (
        _input["velocity_gradient"][1].func
        == _velocity.simple_shear_2d("Y", "X", 5e-6)[1].func
    )
    assert (
        _input["velocity_gradient"][1].keywords
        == _velocity.simple_shear_2d("Y", "X", 5e-6)[1].keywords
    )
    assert _input["locations_initial"] == _io.read_scsv(
        _io.data("specs") / "start.scsv"
    )
    assert _input["timestep"] == 1e9
    assert _input["paths"] is None
    outdir = pl.Path(_io.data("specs") / "out").resolve()
    _output = config["output"]
    assert _output["directory"] == outdir
    np.testing.assert_array_equal(_output["raw_output"], [_core.MineralPhase.olivine])
    np.testing.assert_array_equal(_output["diagnostics"], [_core.MineralPhase.olivine])
    assert _output["anisotropy"] is True
    assert _output["paths"] is None
    assert _output["log_level"] == "DEBUG"
    assert config["parameters"] == {
        "olivine_fraction": 1.0,
        "enstatite_fraction": 0.0,
        "initial_olivine_fabric": _core.MineralFabric.olivine_A,
        "stress_exponent": 1.5,
        "deformation_exponent": 3.5,
        "gbm_mobility": 125,
        "gbs_threshold": 0.3,
        "nucleation_efficiency": 5.0,
        "number_of_grains": 2000,
    }



r"""
#### Simulate crystallographic preferred orientation evolution in polycrystals

---

.. warning::
    **This software is currently in early development (alpha)
    and therefore subject to breaking changes without notice.**

## About

Viscoplastic deformation of minerals, e.g. in Earth's mantle, leads to distinct
signatures in the mineral texture. Many minerals naturally occur in
polycrystalline form, which means that they are composed of many grains with
different volumes and lattice orientations. Preferential alignment of the
average lattice orientation is called crystallographic preferred orientation
(CPO). PyDRex simulates the development and evolution of CPO in deforming
polycrystals, as well as tracking macroscopic finite strain measures.
Currently, the code supports olivine and enstatite mineral phases. The
following features are provided:
- JIT-compiled CPO solver, based on the D-Rex model, which updates the
  polycrystal orientation distribution depending on the macroscopic velocity
  gradients
- `Mineral` class which stores attributes of a distinct mineral phase in the
  polycrystal and its texture snapshots
- Voigt averaging to calculate the average elastic tensor of a textured,
  multiphase polycrystal
- Decomposition of average elastic tensors into components attributed to
  minerals with distinct lattice symmetries
- Crystallographic pole figure visualisation (contouring is a work in progress)
- [work in progress] Texture diagnostics: M-index, bingham average,
  Point-Girdle-Random symmetry, coaxial a.k.a "BA" index, etc.
- [work in progress] Seismic anisotropy diagnostics: % tensorial anisotropy,
  hexagonal symmetry a.k.a transverse isotropy direction, etc.

The core CPO solver is based on the original Fortran 90 implementation by Édouard Kaminski,
which can be [downloaded from this link (~90KB)](http://www.ipgp.fr/~kaminski/web_doudoud/DRex.tar.gz).
The reference papers are [Kaminski & Ribe (2001)](https://doi.org/10.1016/s0012-821x(01)00356-9)
and [Kaminski & Ribe (2004)](https://doi.org/10.1111%2Fj.1365-246x.2004.02308.x),
and an open-access paper which discusses the model is [Fraters & Billen (2021)](https://doi.org/10.1029/2021gc009846).

## Installation

The minimum required Python version is set using `requires-python` in the
[`pyproject.toml`](https://github.com/seismic-anisotropy/PyDRex/blob/main/pyproject.toml) file.
For installation instructions,
see [the README](https://github.com/seismic-anisotropy/PyDRex/blob/main/README.md) file.

## Documentation

The website menu can be used to discover the public API of this package.
Some of the tests are also documented and can serve as usage examples.
Their docstrings can be viewed [in this section](../tests.html).
Documentation is also available from the Python REPL via the `help()` method.

## The D-Rex kinematic CPO model

The D-Rex model is used to compute crystallographic preferred orientation (CPO)
for polycrystals deforming by plastic deformation and dynamic recrystallization.
Polycrystals are discretized into "grains" which represent fractional volumes
of the total crystal that are characterised by a particular crystal lattice
orientation. For numerical efficiency, the number of grains in the model does
not change, and should only be interpreted as an approximation of the number
of (unrecrystallised) physical grains. Dynamic recrystallization is modelled using
statistical expressions which approximate the interaction of each grain with an
effective medium based on the averaged dislocation energy of all other grains. Note that
the model is not suited to situations where static recrystallization processes are
significant.

The primary microphysical mechanism for plastic deformation of the polycrystal
is dislocation creep, which involves dislocation glide ("slip") along symmetry
planes of the mineral and dislocation climb, which allows for dislocations to
annihilate each other so that the number of dislocations reaches a steady-state.
The D-Rex model does not simulate dislocation climb, but implicitly assumes that
the dislocations are in steady-state so that the dislocation density of the
crystal can be described by

$$
ρ ∝ b^{-2} \left(\frac{σ}{μ}\right)^{p}
$$

where $b$ is the length of the Burgers' vector, $σ$ is the stress
and $μ$ is the shear modulus. The value of the exponent $p$ is given by the
`stress_exponent` input parameter. For an overview of available parameters,
see [the `pydrex.mock` source code, for now...]

The effects of dynamic recrystallization are twofold. Grains with a higher than
average dislocation density may be affected by either grain nucleation, which is
the formation of initially small, strain-free sub-grains, or grain boundary
migration, by which process other grains of lower strain energy annex a portion
of its volume. Nucleation occurs mostly in grains oriented favourably for
dislocation glide, and the new grains also grow by grain boundary migration.
If nucleation is too inefficient, the dislocation density in deformation-aligned
grains will remain high and these grains will therefore shrink in volume. On the
other hand, if grain boundaries are too immobile, then nucleated grains will take
longer to grow, reducing the speed of CPO development and re-orientation.
Because nucleated grains are assumed to inherit the orientation of the parent,
they do not affect the model except by reducing the average dislocation density.
A grain boundary mobility parameter of $M^{∗} = 0$ will therefore disable any
recrystallization effects. Finally, the process of grain boundary sliding can
also be included, which simply disallows rotation of grains with very small volume.
This only affects CPO evolution by introducing a latency for the onset of grain
boundary migration in nucleated grains. It also manifests as an upper bound on
texture strength.

## Parameter reference

Model parameters will eventually be provided in a `.toml` file.
For now just pass a dictionary to `config` in the `Mineral.update_orientations` method.
A draft of the input file spec is shown below:

```toml
.. include:: data/specs/spec.toml
```

"""

# Set up the top-level pydrex namespace for convenient usage.
# To keep it clean, we don't want every single symbol here, especially not those from
# `utils` or `visualisation` modules, which should be explicitly imported instead.
import pydrex.axes  # Defines the 'pydrex.polefigure' Axes subclass.
from pydrex.core import (
    DeformationRegime,
    MineralFabric,
    MineralPhase,
    derivatives,
    get_crss,
)
from pydrex.diagnostics import (
    bingham_average,
    coaxial_index,
    elasticity_components,
    finite_strain,
    misorientation_index,
    misorientation_indices,
    smallest_angle,
    symmetry_pgr,
)
from pydrex.geometry import (
    LatticeSystem,
    lambert_equal_area,
    misorientation_angles,
    poles,
    shirley_concentric_squaredisk,
    symmetry_operations,
    to_cartesian,
    to_spherical,
)
from pydrex.io import DEFAULT_PARAMS, data, read_scsv, save_scsv
from pydrex.minerals import (
    OLIVINE_PRIMARY_AXIS,
    OLIVINE_SLIP_SYSTEMS,
    OLIVINE_STIFFNESS,
    Mineral,
    voigt_averages,
)
from pydrex.pathlines import get_pathline
from pydrex.stats import (
    misorientation_hist,
    misorientations_random,
    resample_orientations,
)
from pydrex.tensors import (
    elastic_tensor_to_voigt,
    rotate,
    voigt_decompose,
    voigt_matrix_to_vector,
    voigt_to_elastic_tensor,
    voigt_vector_to_matrix,
)



"""> PyDRex: Custom Matplotlib Axes subclasses."""

import matplotlib as mpl
import matplotlib.axes as mplax
import numpy as np
from matplotlib.projections import register_projection

from pydrex import geometry as _geo
from pydrex import stats as _stats


class PoleFigureAxes(mplax.Axes):
    """Axes class designed for crystallographic pole figures.

    Thin matplotlib Axes wrapper for crystallographic pole figures.

    .. note::
        Projections are not performed automatically using default methods like
        `scatter` or `plot`. To actually plot the pole figures, use `polefigure`.

    """

    name = "pydrex.polefigure"

    def _prep_polefig_axis(self, ref_axes="xz"):
        """Set various options of a matplotlib `Axes` to prepare for a pole figure.

        Use a two-letter string for `ref_axes`.
        These letters will be used for the horizontal and vertical labels, respectively.

        """
        self.set_axis_off()
        self.set_aspect("equal")
        _circle_points = np.linspace(0, np.pi * 2, 100)
        self.plot(
            np.cos(_circle_points),
            np.sin(_circle_points),
            linewidth=0.25,
            color=mpl.rcParams["axes.edgecolor"],
        )
        self.axhline(0, color=mpl.rcParams["grid.color"], alpha=0.5)
        self.text(
            1.0,
            0.4,
            ref_axes[0],
            verticalalignment="center",
            transform=self.transAxes,
            fontsize="x-small",
        )
        self.axvline(0, color=mpl.rcParams["grid.color"], alpha=0.5)

        self.text(
            0.6,
            1.0,
            ref_axes[1],
            horizontalalignment="center",
            transform=self.transAxes,
            fontsize="x-small",
        )

    def polefigure(
        self,
        data,
        density=False,
        ref_axes="xz",
        hkl=[1, 0, 0],
        density_kwargs=None,
        **kwargs,
    ):
        """Plot pole figure of crystallographic texture.

        Args:
        - `data` (array) — Nx3x3 array of orientation matrices
        - `density` (bool, optional) — plot contoured pole figures, False by default
        - `ref_axes` (string, optional) — letters specifying the horizontal and vertical
          axes of the pole figure, and respective labels
        - `hkl` (array, optional) — crystallographic axis (one of the slip
          directions of olivine, i.e. [1, 0, 0], [0, 1, 0] or [0, 0, 1])
        - `density_kwargs` (dict, optional) — keyword arguments to pass to
          `pydrex.stats.point_density` if `density=True`

        Any additional keyword arguments are passed to either `tripcolor` if
        `density=True` or `scatter` if `density=False`

        """
        if density_kwargs is None:
            density_kwargs = {}

        self._prep_polefig_axis(ref_axes=ref_axes)

        if density:
            x, y, z = _stats.point_density(
                *_geo.poles(data, hkl=hkl, ref_axes=ref_axes), **density_kwargs
            )
            return self.tripcolor(
                x.ravel(),
                y.ravel(),
                z.ravel(),
                shading=kwargs.pop("shading", "gouraud"),
                **kwargs,
            )
        else:
            return self.scatter(
                *_geo.lambert_equal_area(
                    *_geo.poles(data, hkl=hkl, ref_axes=ref_axes),
                ),
                s=kwargs.pop("s", 1),
                c=kwargs.pop("c", mpl.rcParams["axes.edgecolor"]),
                marker=kwargs.pop("marker", "."),
                alpha=kwargs.pop("alpha", 0.33),
                zorder=kwargs.pop("zorder", 11),
                **kwargs,
            )


register_projection(PoleFigureAxes)



#!/usr/bin/env python3
import argparse
import warnings
from time import perf_counter

import matplotlib.pyplot as plt
import numpy as np
import vtk
from DRexParam import gridMax, gridMin, size
from numpy.linalg import eigvalsh
from scipy.integrate import solve_ivp
from scipy.interpolate import CloughTocher2DInterpolator, NearestNDInterpolator
from scipy.spatial import Delaunay
from scipy.spatial.transform import Rotation as R


def interpVel(pointCoords, dictGlobals=None):
    pointVel = np.array(
        [
            iVelX(*pointCoords),
            (iVelY(*pointCoords) if len(pointCoords) == 3 else np.nan),
            -iVelZ(*pointCoords),
        ]
    )
    return pointVel[~np.isnan(pointVel)]


def interpVelGrad(pointCoords, dictGlobals=None, ivpIter=False):
    if len(pointCoords) == 2:
        L = np.array(
            [
                [iLxx(*pointCoords), 0, -iLxz(*pointCoords)],
                [0, 0, 0],
                [-iLzx(*pointCoords), 0, iLzz(*pointCoords)],
            ]
        )
    else:
        L = np.array(
            [
                np.hstack((iLxx(pointCoords), iLxy(pointCoords), -iLxz(pointCoords))),
                np.hstack((iLyx(pointCoords), iLyy(pointCoords), -iLyz(pointCoords))),
                np.hstack((-iLzx(pointCoords), -iLzy(pointCoords), iLzz(pointCoords))),
            ]
        )
    assert abs(np.sum(np.diag(L))) < 1e-15, [L, pointCoords]
    if ivpIter:
        return L
    e = (L + np.transpose(L)) / 2
    epsnot = np.amax(np.absolute(eigvalsh(e)))
    return L, epsnot


def pathline(
    currPoint, dictGlobals, ivpMethod, first_step, max_step, t_eval, atol, rtol
):
    def ivpFunc(time, pointCoords, dictGlobals=None):
        if isInside(pointCoords, dictGlobals):
            return interpVel(pointCoords, dictGlobals)
        else:
            return np.zeros(pointCoords.shape)

    def ivpJac(time, pointCoords, dictGlobals=None):
        if isInside(pointCoords, dictGlobals):
            L = interpVelGrad(pointCoords, dictGlobals, True)
            if pointCoords.size == 2:
                return np.array([[L[0, 0], L[0, 2]], [L[2, 0], L[2, 2]]])
            else:
                return L
        else:
            return np.zeros((pointCoords.size, pointCoords.size))

    def maxStrain(time, pointCoords, dictGlobals=None):
        nonlocal eventTime, eventTimePrev, eventStrain, eventStrainPrev, eventFlag
        if eventFlag:
            return (eventStrain if time == eventTime else eventStrainPrev) - 10
        elif isInside(pointCoords, dictGlobals):
            L, epsnot = interpVelGrad(pointCoords, dictGlobals)
            eventStrainPrev = eventStrain
            eventStrain += abs(time - eventTime) * epsnot
            if eventStrain >= 10:
                eventFlag = True
            eventTimePrev = eventTime
            eventTime = time
            return eventStrain - 10
        else:
            return 0

    maxStrain.terminal = True
    eventStrain = eventTime = 0
    eventStrainPrev = eventTimePrev = None
    eventFlag = False
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        sol = solve_ivp(
            ivpFunc,
            [0, -100e6 * 365.25 * 8.64e4],
            currPoint,
            method=ivpMethod,
            first_step=first_step,
            max_step=max_step,
            t_eval=t_eval,
            events=[maxStrain],
            args=(dictGlobals,),
            dense_output=True,
            jac=ivpJac,
            atol=atol,
            rtol=rtol,
        )
    return sol.t, sol.sol


def isInside(point, dictGlobals=None):
    inside = True
    for coord, minBound, maxBound in zip(point, gridMin, gridMax):
        if coord < minBound or coord > maxBound:
            inside = False
            break
    return inside


parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""
A simple plotting script using Matplotlib to generate filled contour plots.
Requires the input file DRexParam.py in the same directory.""",
)
parser.add_argument("input", help="input file (expects a VTK unstructured grid)")
args = parser.parse_args()

vtkReader = vtk.vtkXMLUnstructuredGridReader()
vtkReader.SetFileName(args.input)
vtkReader.Update()
vtkOut = vtkReader.GetOutput()
vtkData = vtkOut.GetPoints().GetData()
coords = np.array(
    [vtkData.GetTuple3(tup) for tup in range(vtkData.GetNumberOfTuples())]
)
if np.amin(coords[:, 1]) == np.amax(coords[:, 1]):
    raise RuntimeError("""Please either provide a 2-D or 3-D dataset.""") from None
elif np.amin(coords[:, 2]) == np.amax(coords[:, 2]):
    dim = 2
    coords[:, 1] = np.amax(coords[:, 1]) - coords[:, 1]
else:
    dim = 3
    coords[:, 2] = np.amax(coords[:, 2]) - coords[:, 2]
vtkScalars = vtkOut.GetPointData().GetScalars("DeformationMechanism")
defMech = np.array(
    [vtkScalars.GetTuple1(tup) for tup in range(vtkScalars.GetNumberOfTuples())]
)
vtkScalars = vtkOut.GetPointData().GetScalars("Velocity")
vel = np.array(
    [vtkScalars.GetTuple3(tup) for tup in range(vtkScalars.GetNumberOfTuples())]
)
vtkArray = vtkOut.GetPointData().GetArray("VelocityGradient")
velGrad = np.array(
    [vtkArray.GetTuple9(tup) for tup in range(vtkArray.GetNumberOfTuples())]
)
del vtkReader, vtkData, vtkScalars, vtkArray

if dim == 2:
    iDefMech = NearestNDInterpolator(coords[:, :-1], defMech)
    tri = Delaunay(coords[:, :-1])
    iVelX = CloughTocher2DInterpolator(tri, vel[:, 0])
    iVelZ = CloughTocher2DInterpolator(tri, vel[:, 1])
    iLxx = CloughTocher2DInterpolator(tri, velGrad[:, 0])
    iLzx = CloughTocher2DInterpolator(tri, velGrad[:, 1])
    iLxz = CloughTocher2DInterpolator(tri, velGrad[:, 3])
    iLzz = CloughTocher2DInterpolator(tri, velGrad[:, 4])
else:
    iDefMech = NearestNDInterpolator(coords, defMech)
    iVelX = NearestNDInterpolator(coords, vel[:, 0])
    iVelY = NearestNDInterpolator(coords, vel[:, 1])
    iVelZ = NearestNDInterpolator(coords, vel[:, 2])
    iLxx = NearestNDInterpolator(coords, velGrad[:, 0])
    iLyx = NearestNDInterpolator(coords, velGrad[:, 1])
    iLzx = NearestNDInterpolator(coords, velGrad[:, 2])
    iLxy = NearestNDInterpolator(coords, velGrad[:, 3])
    iLyy = NearestNDInterpolator(coords, velGrad[:, 4])
    iLzy = NearestNDInterpolator(coords, velGrad[:, 5])
    iLxz = NearestNDInterpolator(coords, velGrad[:, 6])
    iLyz = NearestNDInterpolator(coords, velGrad[:, 7])
    iLzz = NearestNDInterpolator(coords, velGrad[:, 8])
del vtkOut

alt = np.zeros((3, 3, 3))  # \epsilon_{ijk}
for ii in range(3):
    alt[ii % 3, (ii + 1) % 3, (ii + 2) % 3] = 1
    alt[ii % 3, (ii + 2) % 3, (ii + 1) % 3] = -1
# ijkl -> Tensor of indices to form Cijkl from Sij
ijkl = np.array([[0, 5, 4], [5, 1, 3], [4, 3, 2]], dtype="int")
# l1, l2 -> Tensors of indices to form Sij from Cijkl
l1 = np.array([0, 1, 2, 1, 2, 0], dtype="int")
l2 = np.array([0, 1, 2, 2, 0, 1], dtype="int")
# Direction cosine matrix with uniformly distributed rotations.
acs0 = R.random(size, random_state=1).as_matrix()


method = ["RK23", "RK45", "DOP853", "Radau", "BDF", "LSODA"]
plotTime = -1e4 * 365.25 * 8.64e4 * np.arange(10001)
if dim == 3:
    fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    point = [2.4e6, 7e5, 2e4]
else:
    fig, ax = plt.subplots()
    point = [1.1e6, 1e5]
for name in method:
    begin = perf_counter()
    pathTime, pathDense = pathline(
        point,
        None,
        name,
        first_step=1e11,
        max_step=np.inf,
        t_eval=None,
        atol=1e-8,
        rtol=1e-5,
    )
    end = perf_counter()
    if dim == 3:
        ax.plot3D(
            *pathDense(plotTime[plotTime > pathTime[-1]]),
            label=f"{name}: {end - begin:.3f}s",
        )
    else:
        ax.plot(
            *pathDense(plotTime[plotTime > pathTime[-1]]),
            label=f"{name}: {end - begin:.3f}s",
        )
    begin = perf_counter()
    pathTime, pathDense = pathline(
        point,
        None,
        name,
        first_step=None,
        max_step=5e11,
        t_eval=None,
        atol=1e-9,
        rtol=1e-6,
    )
    end = perf_counter()
    if dim == 3:
        ax.plot3D(
            *pathDense(plotTime[plotTime > pathTime[-1]]),
            label=f"{name}: {end - begin:.3f}s",
            linestyle="dashed",
        )
    else:
        ax.plot(
            *pathDense(plotTime[plotTime > pathTime[-1]]),
            label=f"{name}: {end - begin:.3f}s",
            linestyle="dashed",
        )
ax.invert_yaxis()
ax.set_aspect("auto") if dim == 3 else ax.set_aspect("equal")
ax.legend(ncol=6)
plt.show()



"""> PyDRex: Miscellaneous utility methods."""

import os
import platform
import subprocess

import numba as nb
import numpy as np
from matplotlib.collections import PathCollection
from matplotlib.legend_handler import HandlerLine2D, HandlerPathCollection
from matplotlib.pyplot import Line2D
from matplotlib.transforms import ScaledTranslation

from pydrex import logger as _log


def import_proc_pool():
    """Import either `ray.util.multiprocessing.Pool` or `multiprocessing.Pool`.

    Import a process `Pool` object either from Ray of from Python's stdlib.
    Both offer the same API, the Ray implementation will be preferred if available.
    Using the `Pool` provided by Ray allows for distributed memory multiprocessing.

    Returns a tuple containing the `Pool` object and a boolean flag which is `True` if
    Ray is available.

    """
    try:
        from ray.util.multiprocessing import Pool

        has_ray = True
    except ImportError:
        from multiprocessing import Pool

        has_ray = False
    return Pool, has_ray


@nb.njit(fastmath=True)
def strain_increment(dt, velocity_gradient):
    """Calculate strain increment for a given time increment and velocity gradient.

    Returns “tensorial” strain increment ε, which is equal to γ/2 where γ is the
    “(engineering) shear strain” increment.

    """
    return (
        np.abs(dt)
        * np.abs(
            np.linalg.eigvalsh((velocity_gradient + velocity_gradient.transpose()) / 2)
        ).max()
    )


@nb.njit
def apply_gbs(orientations, fractions, gbs_threshold, orientations_prev, n_grains):
    """Apply grain boundary sliding for small grains."""
    mask = fractions < (gbs_threshold / n_grains)
    # _log.debug(
    #     "grain boundary sliding activity (volume percentage): %s",
    #     len(np.nonzero(mask)) / len(fractions),
    # )
    # No rotation: carry over previous orientations.
    orientations[mask, :, :] = orientations_prev[mask, :, :]
    fractions[mask] = gbs_threshold / n_grains
    fractions /= fractions.sum()
    # _log.debug(
    #     "grain volume fractions: median=%e, min=%e, max=%e, sum=%e",
    #     np.median(fractions),
    #     np.min(fractions),
    #     np.max(fractions),
    #     np.sum(fractions),
    # )
    return orientations, fractions


@nb.njit
def extract_vars(y, n_grains):
    """Extract deformation gradient, orientation matrices and grain sizes from y."""
    deformation_gradient = y[:9].reshape((3, 3))
    orientations = y[9 : n_grains * 9 + 9].reshape((n_grains, 3, 3)).clip(-1, 1)
    fractions = y[n_grains * 9 + 9 : n_grains * 10 + 9].clip(0, None)
    fractions /= fractions.sum()
    return deformation_gradient, orientations, fractions


def remove_nans(a):
    """Remove NaN values from array."""
    a = np.asarray(a)
    return a[~np.isnan(a)]


def remove_dim(a, dim):
    """Remove all values corresponding to dimension `dim` from an array.

    Note that a `dim` of 0 refers to the “x” values.

    Examples:

    >>> a = [1, 2, 3]
    >>> remove_dim(a, 0)
    array([2, 3])
    >>> remove_dim(a, 1)
    array([1, 3])
    >>> remove_dim(a, 2)
    array([1, 2])

    >>> a = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    >>> remove_dim(a, 0)
    array([[5, 6],
           [8, 9]])
    >>> remove_dim(a, 1)
    array([[1, 3],
           [7, 9]])
    >>> remove_dim(a, 2)
    array([[1, 2],
           [4, 5]])

    """
    _a = np.asarray(a)
    for i, _ in enumerate(_a.shape):
        _a = np.delete(_a, [dim], axis=i)
    return _a


def add_dim(a, dim, val=0):
    """Add entries of `val` corresponding to dimension `dim` to an array.

    Note that a `dim` of 0 refers to the “x” values.

    Examples:

    >>> a = [1, 2]
    >>> add_dim(a, 0)
    array([0, 1, 2])
    >>> add_dim(a, 1)
    array([1, 0, 2])
    >>> add_dim(a, 2)
    array([1, 2, 0])

    >>> add_dim([1.0, 2.0], 2)
    array([1., 2., 0.])

    >>> a = [[1, 2], [3, 4]]
    >>> add_dim(a, 0)
    array([[0, 0, 0],
           [0, 1, 2],
           [0, 3, 4]])
    >>> add_dim(a, 1)
    array([[1, 0, 2],
           [0, 0, 0],
           [3, 0, 4]])
    >>> add_dim(a, 2)
    array([[1, 2, 0],
           [3, 4, 0],
           [0, 0, 0]])

    """
    _a = np.asarray(a)
    for i, _ in enumerate(_a.shape):
        _a = np.insert(_a, [dim], 0, axis=i)
    return _a


def default_ncpus():
    """Get a safe default number of CPUs available for multiprocessing.

    On Linux platforms that support it, the method `os.sched_getaffinity()` is used.
    On Mac OS, the command `sysctl -n hw.ncpu` is used.
    On Windows, the environment variable `NUMBER_OF_PROCESSORS` is queried.
    If any of these fail, a fallback of 1 is used and a warning is logged.

    """
    try:
        match platform.system():
            case "Linux":
                return len(os.sched_getaffinity(0)) - 1  # May raise AttributeError.
            case "Darwin":
                # May raise CalledProcessError.
                out = subprocess.run(
                    ["sysctl", "-n", "hw.ncpu"], capture_output=True, check=True
                )
                return int(out.stdout.strip()) - 1
            case "Windows":
                return int(os.environ["NUMBER_OF_PROCESSORS"]) - 1
    except AttributeError or subprocess.CalledProcessError or KeyError:
        return 1


def diff_like(a):
    """Get forward difference of 2D array `a`, with repeated last elements.

    The repeated last elements ensure that output and input arrays have equal shape.

    Examples:

    >>> diff_like(np.array([1, 2, 3, 4, 5]))
    array([[1, 1, 1, 1, 1]])

    >>> diff_like(np.array([[1, 2, 3, 4, 5], [1, 3, 6, 9, 10]]))
    array([[1, 1, 1, 1, 1],
           [2, 3, 3, 1, 1]])

    >>> diff_like(np.array([[1, 2, 3, 4, 5], [1, 3, 6, 9, 10], [1, 0, 0, 0, np.inf]]))
    array([[ 1.,  1.,  1.,  1.,  1.],
           [ 2.,  3.,  3.,  1.,  1.],
           [-1.,  0.,  0., inf, nan]])

    """
    a2 = np.atleast_2d(a)
    return np.diff(
        a2, append=np.reshape(a2[:, -1] + (a2[:, -1] - a2[:, -2]), (a2.shape[0], 1))
    )


def angle_fse_simpleshear(strain):
    """Get angle of FSE long axis anticlockwise from the X axis in simple shear."""
    return np.rad2deg(np.arctan(np.sqrt(strain**2 + 1) + strain))


def lag_2d_corner_flow(θ):
    """Get predicted grain orientation lag for 2D corner flow.

    See eq. 11 in [Kaminski & Ribe (2002)](https://doi.org/10.1029/2001GC000222).

    """
    _θ = np.ma.masked_less(θ, 1e-15)
    return (_θ * (_θ**2 + np.cos(_θ) ** 2)) / (
        np.tan(_θ) * (_θ**2 + np.cos(_θ) ** 2 - _θ * np.sin(2 * _θ))
    )


@nb.njit(fastmath=True)
def quat_product(q1, q2):
    """Quaternion product, q1, q2 and output are in scalar-last (x,y,z,w) format."""
    return [
        *q1[-1] * q2[:3] + q2[-1] * q1[:3] + np.cross(q1[:3], q1[:3]),
        q1[-1] * q2[-1] - np.dot(q1[:3], q2[:3]),
    ]


def redraw_legend(ax, fig=None, legendax=None, remove_all=True, **kwargs):
    """Redraw legend on matplotlib axis or figure.

    Transparency is removed from legend symbols.
    If `fig` is not None and `remove_all` is True,
    all legends are first removed from the parent figure.
    Optional keyword arguments are passed to `matplotlib.axes.Axes.legend` by default,
    or `matplotlib.figure.Figure.legend` if `fig` is not None.

    If `legendax` is not None, the axis legend will be redrawn using the `legendax` axes
    instead of taking up space in the original axes. This option requires `fig=None`.

    .. warning::
        Note that if `fig` is not `None`, the legend may be cropped from the saved
        figure due to a Matplotlib bug. In this case, it is required to add the
        arguments `bbox_extra_artists=(legend,)` and `bbox_inches="tight"` to `savefig`,
        where `legend` is the object returned by this function. To prevent the legend
        from consuming axes/subplot space, it is further required to add the lines:
        `legend.set_in_layout(False)`, `fig.canvas.draw()`, `legend.set_layout(True)`
        and `fig.set_layout_engine("none")` before saving the figure.

    """
    handler_map = {
        PathCollection: HandlerPathCollection(
            update_func=_remove_legend_symbol_transparency
        ),
        Line2D: HandlerLine2D(update_func=_remove_legend_symbol_transparency),
    }
    if fig is None:
        legend = ax.get_legend()
        if legend is not None:
            handles, labels = ax.get_legend_handles_labels()
            legend.remove()
        if legendax is not None:
            legendax.axis("off")
            return legendax.legend(handles, labels, handler_map=handler_map, **kwargs)
        return ax.legend(handler_map=handler_map, **kwargs)
    else:
        if legendax is not None:
            _log.warning("ignoring `legendax` argument which requires `fig=None`")
        for legend in fig.legends:
            if legend is not None:
                legend.remove()
        if remove_all:
            for ax in fig.axes:
                legend = ax.get_legend()
                if legend is not None:
                    legend.remove()
        return fig.legend(handler_map=handler_map, **kwargs)


def add_subplot_labels(
    mosaic, labelmap=None, loc="left", fontsize="medium", internal=False, **kwargs
):
    """Add subplot labels to axes mosaic.

    Use `labelmap` to specify a dictionary that maps keys in `mosaic` to subplot labels.
    If `labelmap` is None, the keys in `axs` will be used as the labels by default.

    If `internal` is `False` (default), the axes titles will be used.
    Otherwise, internal labels will be drawn with `ax.text`,
    in which case `loc` must be a tuple of floats.

    Any axes in `axs` corresponding to the special key `legend` are skipped.

    """
    for txt, ax in mosaic.items():
        if txt.lower() == "legend":
            continue
        _txt = labelmap[txt] if labelmap is not None else txt
        if internal:
            trans = ScaledTranslation(10 / 72, -5 / 72, ax.figure.dpi_scale_trans)
            if isinstance(loc, str):
                raise ValueError(
                    "'loc' argument must be a sequence of float when 'internal' is 'True'"
                )
            ax.text(
                *loc,
                _txt,
                transform=ax.transAxes + trans,
                fontsize=fontsize,
                bbox={
                    "facecolor": (1.0, 1.0, 1.0, 0.3),
                    "edgecolor": "none",
                    "pad": 3.0,
                },
            )
        else:
            ax.set_title(_txt, loc=loc, fontsize=fontsize, **kwargs)


def _remove_legend_symbol_transparency(handle, orig):
    """Remove transparency from symbols used in a Matplotlib legend."""
    # https://stackoverflow.com/a/59629242/12519962
    handle.update_from(orig)
    handle.set_alpha(1)



"""> PyDRex: logger settings and boilerplate.

Python's logging module is weird and its methods don't allow us to specify
which logger to use, so just using `logging.debug` for example always uses
the "root" logger, which spams a bunch of messages from other imports/modules.
Instead, the methods in this module are thin wrappers that use custom
logging objects (`pydrex.logger.LOGGER` and `pydrex.logger.CONSOLE_LOGGER`).
The method `quiet_aliens` can be invoked to suppress most messages
from third-party modules, except critical errors and warnings from Numba.

By default, PyDRex emits INFO level messages to the console.
This can be changed globally by setting the new level with `CONSOLE_LOGGER.setLevel`:

```python
from pydrex import logger as _log
_log.info("this message will be printed to the console")

_log.CONSOLE_LOGGER.setLevel("ERROR")
_log.info("this message will NOT be printed to the console")
_log.error("this message will be printed to the console")
```

To change the console logging level for a particular local context,
use the `handler_level` context manager:

```python
_log.CONSOLE_LOGGER.setLevel("INFO")
_log.info("this message will be printed to the console")

with handler_level("ERROR"):
    _log.info("this message will NOT be printed to the console")

_log.info("this message will be printed to the console")
```

To save debug logs to a file, the `logfile_enable` context manager is recommended.
Always use the old printf style formatting for log messages, not fstrings,
otherwise compute time will be wasted on string conversions when logging is disabled:

```python
_log.quiet_aliens()  # Suppress third-party log messages except CRITICAL from Numba.
with _log.logfile_enable("my_log_file.log"):  # Overwrite existing file unless mode="a".
    value = 42
    _log.critical("critical error with value: %s", value)
    _log.error("runtime error with value: %s", value)
    _log.warning("warning with value: %s", value)
    _log.info("information message with value: %s", value)
    _log.debug("verbose debugging message with value: %s", value)
    ... # Construct Minerals, update orientations, etc.

```

"""

import contextlib as cl
import functools as ft
import logging
import sys

import numpy as np

from pydrex import io as _io

np.set_printoptions(
    formatter={"float_kind": np.format_float_scientific},
    linewidth=1000,
)
np.set_string_function(ft.partial(np.array2string, separator=", "), repr=False)


class ConsoleFormatter(logging.Formatter):
    """Log formatter that uses terminal color codes."""

    def colorfmt(self, code):
        return (
            f"\033[{code}m%(levelname)s [%(asctime)s]\033[m"
            + " \033[1m%(name)s:\033[m %(message)s"
        )

    def format(self, record):
        format_specs = {
            logging.CRITICAL: self.colorfmt("1;31"),
            logging.ERROR: self.colorfmt("31"),
            logging.INFO: self.colorfmt("32"),
            logging.WARNING: self.colorfmt("33"),
            logging.DEBUG: self.colorfmt("34"),
        }
        self._style._fmt = format_specs.get(record.levelno)
        return super().format(record)


# To create a new logger we use getLogger as recommended by the logging docs.
LOGGER = logging.getLogger("pydrex")
# To allow for multiple handlers at different levels, default level must be DEBUG.
LOGGER.setLevel(logging.DEBUG)
# Set up console handler.
CONSOLE_LOGGER = logging.StreamHandler()
CONSOLE_LOGGER.setFormatter(ConsoleFormatter(datefmt="%H:%M"))
CONSOLE_LOGGER.setLevel(logging.INFO)
# Turn on console logger by default.
LOGGER.addHandler(CONSOLE_LOGGER)


def handle_exception(exec_type, exec_value, exec_traceback):
    # Ignore KeyboardInterrupt so ^C (ctrl + C) works as expected.
    if issubclass(exec_type, KeyboardInterrupt):
        sys.__excepthook__(exec_type, exec_value, exec_traceback)
        return
    # Send other exceptions to the logger.
    LOGGER.exception(
        "uncaught exception", exc_info=(exec_type, exec_value, exec_traceback)
    )


# Make our logger handle uncaught exceptions.
sys.excepthook = handle_exception


@cl.contextmanager
def handler_level(level, handler=CONSOLE_LOGGER):
    """Set logging handler level for current context.

    Args:
    - `level` (string) — logging level name e.g. "DEBUG", "ERROR", etc.
      See Python's logging module for details.
    - `handler` (optional, `logging.Handler`) — alternative handler to control instead
      of the default, `CONSOLE_LOGGER`.

    """
    default_level = handler.level
    handler.setLevel(level)
    yield
    handler.setLevel(default_level)


@cl.contextmanager
def logfile_enable(path, level=logging.DEBUG, mode="w"):
    """Enable logging to a file at `path` with given `level`."""
    logger_file = logging.FileHandler(_io.resolve_path(path), mode=mode)
    logger_file.setFormatter(
        logging.Formatter(
            "%(levelname)s [%(asctime)s] %(name)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger_file.setLevel(level)
    LOGGER.addHandler(logger_file)
    yield
    logger_file.close()


def critical(msg, *args, **kwargs):
    """Log a CRITICAL message in PyDRex."""
    LOGGER.critical(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    """Log an ERROR message in PyDRex."""
    LOGGER.error(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    """Log a WARNING message in PyDRex."""
    LOGGER.warning(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    """Log an INFO message in PyDRex."""
    LOGGER.info(msg, *args, **kwargs)


def debug(msg, *args, **kwargs):
    """Log a DEBUG message in PyDRex."""
    LOGGER.debug(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):
    """Log a message with level ERROR but retain exception information.

    This function should only be called from an exception handler.

    """
    LOGGER.exception(msg, *args, **kwargs)


def quiet_aliens():
    """Restrict alien loggers 👽 because I'm trying to find MY bugs, thanks."""
    # Only allow warnings or above from root logger.
    logging.getLogger().setLevel(logging.WARNING)
    # Only allow critical stuff from other things.
    for name in logging.Logger.manager.loggerDict.keys():
        if name != "pydrex":
            logging.getLogger(name).setLevel(logging.CRITICAL)
    # Numba is not in the list for some reason, I guess we can leave warnings.
    logging.getLogger("numba").setLevel(logging.WARNING)



"""> PyDRex: tests for the SCSV plain text file format."""

import sys
import tempfile

import numpy as np
import pytest
from numpy import testing as nt

from pydrex import exceptions as _err
from pydrex import io as _io
from pydrex import logger as _log


def test_validate_schema(console_handler):
    """Test SCSV schema validation."""
    schema_nofill = {
        "delimiter": ",",
        "missing": "-",
        "fields": [{"name": "nofill", "type": "float"}],
    }
    schema_nomissing = {
        "delimiter": ",",
        "fields": [{"name": "nomissing", "type": "float", "fill": "NaN"}],
    }
    schema_nofields = {"delimiter": ",", "missing": "-"}
    schema_badfieldname = {
        "delimiter": ",",
        "missing": "-",
        "fields": [{"name": "bad name", "type": "float", "fill": "NaN"}],
    }
    schema_delimiter_eq_missing = {
        "delimiter": ",",
        "missing": ",",
        "fields": [{"name": "baddelim", "type": "float", "fill": "NaN"}],
    }
    schema_delimiter_in_missing = {
        "delimiter": ",",
        "missing": "-,",
        "fields": [{"name": "baddelim", "type": "float", "fill": "NaN"}],
    }
    schema_long_delimiter = {
        "delimiter": ",,",
        "missing": "-",
        "fields": [{"name": "baddelim", "type": "float", "fill": "NaN"}],
    }

    # NOTE: NamedTemporaryFile() already opens the file.
    # Attempting to open it again will cause a crash on Windows so close the file first.
    with _log.handler_level("CRITICAL", console_handler):
        with pytest.raises(_err.SCSVError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_nofill, [[0.1]])
        with pytest.raises(_err.SCSVError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_nomissing, [[0.1]])
        with pytest.raises(_err.SCSVError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_nofields, [[0.1]])
        with pytest.raises(_err.SCSVError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_badfieldname, [[0.1]])
        with pytest.raises(_err.SCSVError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_delimiter_eq_missing, [[0.1]])
        with pytest.raises(_err.SCSVError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_delimiter_in_missing, [[0.1]])
        # CSV module already raises a TypeError on long delimiters.
        with pytest.raises(TypeError):
            temp = tempfile.NamedTemporaryFile()
            temp.close()
            _io.save_scsv(temp.name, schema_long_delimiter, [[0.1]])


def test_read_specfile():
    """Test SCSV spec file parsing."""
    data = _io.read_scsv(_io.data("specs") / "spec.scsv")
    assert data._fields == (
        "first_column",
        "second_column",
        "third_column",
        "float_column",
        "bool_column",
        "complex_column",
    )
    nt.assert_equal(data.first_column, ["s1", "MISSING", "s3"])
    nt.assert_equal(data.second_column, ["A", "B, b", ""])
    nt.assert_equal(data.third_column, [999999, 10, 1])
    nt.assert_equal(data.float_column, [1.1, np.nan, 1.0])
    nt.assert_equal(data.bool_column, [True, False, True])
    nt.assert_equal(data.complex_column, [0.1 + 0 * 1j, np.nan + 0 * 1j, 1.0 + 1 * 1j])


@pytest.mark.skipif(sys.platform == "win32", reason="Items are not equal")
def test_save_specfile(outdir, named_tempfile_kwargs):
    """Test SCSV spec file reproduction."""
    schema = {
        "delimiter": ",",
        "missing": "-",
        "fields": [
            {
                "name": "first_column",
                "type": "string",
                "fill": "MISSING",
                "unit": "percent",
            },
            {"name": "second_column"},
            {"name": "third_column", "type": "integer", "fill": "999999"},
            {"name": "float_column", "type": "float", "fill": "NaN"},
            {"name": "bool_column", "type": "boolean"},
            {"name": "complex_column", "type": "complex", "fill": "NaN"},
        ],
    }
    schema_alt = {
        "delimiter": ",",
        "missing": "-",
        "fields": [
            {
                "name": "first_column",
                "type": "string",
                "unit": "percent",
            },
            {"name": "second_column"},
            {"name": "third_column", "type": "integer", "fill": "999991"},
            {"name": "float_column", "type": "float", "fill": "0.0"},
            {"name": "bool_column", "type": "boolean"},
            {"name": "complex_column", "type": "complex", "fill": "NaN"},
        ],
    }

    data = [
        ["s1", "MISSING", "s3"],
        ["A", "B, b", ""],
        [999999, 10, 1],
        [1.1, np.nan, 1.0],
        [True, False, True],
        [0.1 + 0 * 1j, np.nan + 0 * 1j, 1.0 + 1 * 1j],
    ]
    data_alt = [
        ["s1", "", "s3"],
        ["A", "B, b", ""],
        [999991, 10, 1],
        [1.1, 0.0, 1.0],
        [True, False, True],
        [0.1 + 0 * 1j, np.nan + 0 * 1j, 1.0 + 1 * 1j],
    ]

    # The test writes two variants of the file, with identical CSV contents but
    # different YAML header specs. Contents after the header must match.
    if outdir is not None:
        _io.save_scsv(f"{outdir}/spec_out.scsv", schema, data)
        _io.save_scsv(f"{outdir}/spec_out_alt.scsv", schema_alt, data_alt)

    # https://docs.python.org/3/library/tempfile.html#tempfile.NamedTemporaryFile
    temp = tempfile.NamedTemporaryFile(**named_tempfile_kwargs)
    temp_alt = tempfile.NamedTemporaryFile(**named_tempfile_kwargs)
    _io.save_scsv(temp.name, schema, data)
    _io.save_scsv(temp_alt.name, schema_alt, data_alt)
    raw_read = []
    raw_read_alt = []
    with open(temp.name) as stream:
        raw_read = stream.readlines()[23:]  # Extra spec for first column 'fill' value.
    with open(temp_alt.name) as stream:
        raw_read_alt = stream.readlines()[22:]
    _log.debug("\n  first file: %s\n  second file: %s", raw_read, raw_read_alt)
    nt.assert_equal(raw_read, raw_read_alt)


def test_read_Kaminski2002():
    data = _io.read_scsv(_io.data("thirdparty") / "Kaminski2002_ISAtime.scsv")
    assert data._fields == ("time_ISA", "vorticity")
    # fmt: off
    nt.assert_equal(
        data.time_ISA,
        np.array(
            [2.48, 2.50, 2.55, 2.78, 3.07, 3.58, 4.00, 4.88, 4.01, 3.79,
             3.72, 3.66, 3.71, 4.22, 4.73, 3.45, 1.77, 0.51]
        ),
    )
    nt.assert_equal(
        data.vorticity,
        np.array(
            [0.05, 0.10, 0.20, 0.30, 0.35, 0.40, 0.45, 0.50, 0.55, 0.60,
             0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95, 1.00]
        ),
    )
    # fmt: on


def test_save_scsv_errors(named_tempfile_kwargs):
    """Check that we raise errors when attempting to write bad SCSV data."""
    schema = {
        "delimiter": ",",
        "missing": "-",
        "fields": [
            {
                "name": "foo",
                "type": "integer",
                "fill": 999999,
            }
        ],
    }
    # https://docs.python.org/3/library/tempfile.html#tempfile.NamedTemporaryFile
    temp = tempfile.NamedTemporaryFile(**named_tempfile_kwargs)
    with pytest.raises(_err.SCSVError):
        foo = [1, 5, 0.2]
        _io.save_scsv(temp.name, schema, [foo])
        foo = [1, "foo"]
        _io.save_scsv(temp.name, schema, [foo])
        foo = ["foo"]
        _io.save_scsv(temp.name, schema, [foo])
        foo = [True]
        _io.save_scsv(temp.name, schema, [foo])
        foo = [1, 2, 3]
        bar = [1, 2, 3]
        _io.save_scsv(temp.name, schema, [foo, bar])


def test_read_Kaminski2004():
    data = _io.read_scsv(_io.data("thirdparty") / "Kaminski2004_AaxisDynamicShear.scsv")
    assert data._fields == ("time", "meanA_X0", "meanA_X02", "meanA_X04")
    # fmt: off
    nt.assert_equal(
        data.time,
        np.array(
            [-0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9,
             1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0,
             2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0, 3.1,
             3.2, 3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4.0, 4.1, 4.2,
             4.3, 4.4, 4.5, 4.6, 4.7, 4.8, 4.9, 5.0]
        ),
    )
    nt.assert_equal(
        data.meanA_X02,
        np.array(
            [-0.54, -0.54, -0.27, 0.13, 0.94, 2.82, 5.37, 9.53, 14.77, 20.40,
             26.58, 32.89, 39.73, 47.25, 53.69, 58.66, 60.81, 60.81, 59.73, 58.52, 58.12,
             56.64, 54.09, 53.69, 55.57, 57.05, 58.66, 60.54, 60.81, 61.21, 61.21, 61.61,
             61.48, 61.61, 61.61, 61.61, 61.21, 61.21, 61.07, 60.81, 60.81, 60.54, 60.00,
             59.60, 59.33, 58.52, 58.12, 57.85, 57.45, 57.05, 57.05]
        ),
    )
    # fmt: on


def test_read_Skemer2016():
    data = _io.read_scsv(_io.data("thirdparty") / "Skemer2016_ShearStrainAngles.scsv")
    assert data._fields == (
        "study",
        "sample_id",
        "shear_strain",
        "angle",
        "fabric",
        "M_index",
    )
    # fmt: off
    nt.assert_equal(
        data.study,
        ["Z&K 1200 C", "Z&K 1200 C", "Z&K 1200 C", "Z&K 1200 C", "Z&K 1200 C",
         "Z&K 1300 C", "Z&K 1300 C", "Z&K 1300 C", "Z&K 1300 C", "Z&K 1300 C",
         "Z&K 1300 C", "Z&K 1300 C", "Bystricky 2000", "Bystricky 2000", "Bystricky 2000",
         "Bystricky 2000", "Warren 2008", "Warren 2008", "Warren 2008", "Warren 2008",
         "Warren 2008", "Warren 2008", "Warren 2008", "Warren 2008", "Warren 2008",
         "Skemer 2011", "Skemer 2011", "Skemer 2011", "Webber 2010", "Webber 2010",
         "Webber 2010", "Webber 2010", "Webber 2010", "Webber 2010", "Webber 2010",
         "Katayama 2004", "Katayama 2004", "Katayama 2004", "Katayama 2004", "Katayama 2004",
         "Katayama 2004", "Skemer 2010", "Skemer 2010", "Skemer 2010", "Skemer 2010",
         "Skemer 2010", "Jung 2006", "Jung 2006", "Jung 2006", "Jung 2006",
         "Jung 2006", "Jung 2006", "Hansen 2014", "Hansen 2014", "Hansen 2014",
         "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014",
         "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014",
         "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014",
         "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014", "Hansen 2014",
         "Hansen 2014", "Hansen 2014", "Hansen 2014", "H&W 2015", "H&W 2015",
         "H&W 2015", "H&W 2015", "H&W 2015", "H&W 2015", "H&W 2015",
         "H&W 2015", "H&W 2015", "H&W 2015", "H&W 2015", "H&W 2015",
         "H&W 2015", "H&W 2015", "H&W 2015", "H&W 2015"],
    )
    nt.assert_equal(
        data.sample_id,
        ["PI-148", "PI-150", "PI-154", "PI-158", "PI-284", "MIT-5", "MIT-20", "MIT-6",
         "MIT-21", "MIT-17", "MIT-18", "MIT-19", "", "", "", "", "3923J01",
         "3923J11", "3923J13", "3923J14", "3924J06", "3924J09a", "3924J09b", "3924J08",
         "3924J10", "PIP-20", "PIP-21", "PT-484", "", "", "", "", "", "", "", "GA10",
         "GA38", "GA23", "GA12", "GA45", "GA25", "3925G08", "3925G05", "3925G02",
         "3925G01", "PT-248_4", "JK43", "JK21B", "JK26", "JK11", "JK18", "JK23",
         "PT0535", "PT0248_1", "", "PT0655", "PT0248_2", "PI-284", "PT0248_3", "PT0503_5",
         "PT0248_4", "PT0248_5", "PT0503_4", "PT0640", "PT0503_3", "", "PT0494",
         "PT0538", "PT0503_2", "PT0541", "PT0503_1", "PT0633", "PT0552", "PT0505",
         "PT0651", "PT0499", "PT619", "PT0484", "3923J06", "3923J07", "3923J09",
         "3923J08", "3923J13", "3923J12", "3924J06", "3924J05", "JP13-D07", "JP13-D06",
         "3924J03a", "3924J03b", "3924J09a", "3924J09b", "3924J08", "3924J07"],
    )
    nt.assert_equal(
        data.shear_strain,
        np.array(
            [
                17, 30, 45, 65, 110, 11, 7, 65, 58, 100,
                115, 150, 50, 200, 400, 500, 0, 65, 118, 131, 258, 386,
                386, 525, 168, 120, 180, 350, 0, 25, 100, 130, 168, 330,
                330, 60, 60, 80, 120, 260, 630, 60, 150, 1000, 1000, 290,
                120, 400, 100, 110, 120, 400, 0, 30, 50, 100, 100, 110,
                210, 270, 290, 390, 390, 400, 490, 500, 590, 680, 680, 760,
                820, 870, 880, 1020, 1060, 1090, 1420, 1870, 32, 32, 81, 81,
                118, 118, 258, 258, 286, 286, 337, 337, 386, 386, 525, 525
            ]
        ),
    )
    nt.assert_equal(
        data.angle,
        np.array(
            [
                43, 37, 38, 24, 20, 36, 28, 18, 10, 5,
                10, 0, 0, 0, 0, 0, 62, 37, 49, 61, 4, 11, 0, 1, 33, 55,
                53, 45, 55, 35, 47, 29, 37, 47, 45, -49, -26, -10, -35,
                -10, -15, -52, -30, -14, -11, 0, 26, 15, 27, 25, 16, 36,
                -71, -78, 71, 29, 39, 24, 12, 7.3, -3.7, -1.3, -6.1, 0.4,
                -8.6, 5.4, 0.6, -5.5, -4.3, -8.4, -1.9, -0.9, -8.9, 2.8,
                -2, -1.5, -4.1, -5.8, 48, 51, 44, 35, 35, 40, 1, 15, 25,
                28, 28, 39, 1, 8, 4, 11
            ]
        ),
    )
    nt.assert_equal(
        data.fabric,
        [
            "A", "A", "A", "A", "A", "A", "A", "A", "A", "A", "A", "A",
            "D", "D", "D", "D", "A", "A", "A", "A", "A", "A", "A", "A",
            "A", "A", "A", "A", "A", "A", "A", "A", "A", "A", "A", "E",
            "E", "E", "E", "E", "E", "E", "E", "E", "E", "D", "A", "B",
            "C", "C", "C", "C", "D", "D", "D", "D", "D", "D", "D", "D",
            "D", "D", "D", "D", "D", "A", "A", "A", "A", "A", "A", "A",
            "A", "A", "A", "A", "A", "A", "A", "A", "A", "A", "A", "A",
            "A", "A", "A", "A", "A", "A", "A", "A", "A", "A",
        ],
    )
    nt.assert_equal(
        data.M_index,
        np.array(
            [
                np.nan, np.nan, np.nan, np.nan, 0.09, np.nan, np.nan,
                np.nan, np.nan, np.nan, np.nan, np.nan, 0.05, np.nan,
                np.nan, 0.17, 0.08, 0.12, 0.16, 0.2, 0.17, 0.06, 0.13,
                0.14, 0.16, np.nan, np.nan, np.nan, np.nan, np.nan,
                np.nan, np.nan, np.nan, np.nan, np.nan, 0.17, np.nan,
                np.nan, 0.28, np.nan, 0.44, 0.03, 0.1, 0.47, 0.37,
                0.36, 0.06, 0.21, 0.21, 0.23, 0.27, 0.1, 0.01, 0.06,
                0.05, 0.06, 0.06, 0.09, 0.2, 0.26, 0.36, 0.35, 0.34,
                0.36, 0.45, 0.17, 0.4, 0.46, 0.46, 0.44, 0.35, 0.39,
                0.4, 0.69, 0.55, 0.61, 0.51, 0.51, 0.08, 0.2, 0.16,
                0.12, 0.12, 0.17, 0.18, 0.13, 0.15, 0.09, 0.19, 0.27,
                0.12, 0.18, 0.14, 0.26,
            ]
        ),
    )



"""> PyDRex: Functions for pathline construction."""

import numpy as np
from scipy import integrate as si

from pydrex import logger as _log
from pydrex import utils as _utils


def get_pathline(
    final_location,
    get_velocity,
    get_velocity_gradient,
    min_coords,
    max_coords,
    max_strain,
    regular_steps=None,
    **kwargs,
):
    """Determine the pathline for a particle in a steady state flow.

    The pathline will terminate at the given `final_location` and follow a curve
    determined by the velocity gradient. It works for both 2D (rectangular) and 3D
    (orthopiped¹) domains, so long as the provided callables expect/return arrays of the
    appropriate dimension.

    .. note::
        The pathline is calculated backwards in time (t < 0) from the given endpoint.
        Therefore, the returned position callable should be evaluated at negative times.

    Args:
    - `final_location` (array) — coordinates of the final location
    - `get_velocity` (callable) — returns velocity vector at a point
    - `get_velocity_gradient` (callable) — returns velocity gradient matrix at a point
    - `min_coords` (array) — lower bound coordinates of the box
    - `max_coords` (array) — upper bound coordinates of the box
    - `max_strain` (float) — target strain (given as “tensorial” strain ε) at the final
      location, useful if the pathline never inflows into the domain (the pathline will
      only be traced backwards until a strain of 0 is reached, unless a domain boundary
      is reached first)
    - `regular_steps` (float, optional) — number of time steps to use for regular
      resampling between the start (t << 0) and end (t <= 0) of the pathline
      (if `None`, which is the default, then the timestamps obtained from
      `scipy.integrate.solve_ivp` are returned instead)

    Optional keyword arguments will be passed to `scipy.integrate.solve_ivp`. However,
    some of the arguments to the `solve_ivp` call may not be modified, and a warning
    will be raised if they are provided.

    Returns a tuple containing the time points and an interpolant that can be used
    to evaluate the pathline position (see `scipy.integrate.OdeSolution`).

    ¹An “orthopiped” is a 3D rectangle (called a “box” when we are in a hurry), see
    <https://www.whatistoday.net/2020/04/cuboid-dilemma.html>.

    """

    def _terminate(
        time, point, get_velocity, get_velocity_gradient, min_coords, max_coords
    ):
        # Track “previous” (last seen) timestamp and total strain value.
        nonlocal _time_prev, _strain

        if _is_inside(point, min_coords, max_coords):
            dε = _utils.strain_increment(
                time - _time_prev, get_velocity_gradient(np.nan, point)
            )
            if time > _time_prev:  # Timestamps jump around for SciPy to find the root.
                _strain += dε
            else:  # Subtract strain increment because we are going backwards in time.
                _strain -= dε
            _time_prev = time
            return _strain
        # If we are outside the domain, always terminate.
        return 0

    _terminate.terminal = True
    _strain = max_strain
    _time_prev = 0
    _event_flag = False

    # Illegal keyword args, check the call below. Remove them and warn about it.
    for key in ("events", "jac", "dense_output", "args"):
        try:
            kwargs.pop(key)
        except KeyError:
            continue
        else:
            _log.warning("ignoring illegal keyword argument: %s", key)

    # We don't want to stop at a particular time,
    # so integrate time for 100 Myr, in seconds (“forever”).
    path = si.solve_ivp(
        _ivp_func,
        [0, -100e6 * 365.25 * 8.64e4],
        final_location,
        method=kwargs.pop("method", "LSODA"),
        events=[_terminate],
        args=(get_velocity, get_velocity_gradient, min_coords, max_coords),
        dense_output=True,
        jac=_ivp_jac,
        atol=kwargs.pop("atol", 1e-8),
        rtol=kwargs.pop("rtol", 1e-5),
        **kwargs,
    )
    _log.info(
        "calculated pathline from %s (t = %e) to %s (t = %e)",
        path.sol(path.t[0]),
        path.t[0],
        path.sol(path.t[-1]),
        path.t[-1],
    )

    if regular_steps is None:
        return path.t[::-1], path.sol
    else:
        return np.linspace(path.t[-1], path.t[0], regular_steps + 1), path.sol


def _ivp_func(time, point, get_velocity, get_velocity_gradient, min_coords, max_coords):
    """Internal use only, must have the same signature as `get_pathline`."""
    if _is_inside(point, min_coords, max_coords):
        return get_velocity(np.nan, point)
    return np.zeros_like(point)


def _ivp_jac(time, point, get_velocity, get_velocity_gradient, min_coords, max_coords):
    """Internal use only, must have the same signature as `_ivp_func`."""
    if _is_inside(point, min_coords, max_coords):
        return get_velocity_gradient(np.nan, point)
    return np.zeros((np.array(point).size,) * 2)


def _is_inside(point, min_coords, max_coords):
    """Check if the point lies within the numerical domain."""
    assert np.array(point).size == len(min_coords) == len(max_coords)
    if np.any(np.array(point) < min_coords) or np.any(np.array(point) > max_coords):
        return False
    return True



"""> PyDRex: Run doctests for all submodules."""

import doctest
import importlib
import os
import pkgutil

import numpy as np
import pytest

import pydrex
from pydrex import logger as _log
from pydrex.exceptions import Error


def _get_submodule_list():
    # Reset NumPy print options because doctests are just string matches, and typing out
    # so many significant digits in doctests is annoying.
    np.set_printoptions()
    np.set_string_function(None)
    modules = ["pydrex." + m.name for m in pkgutil.iter_modules(pydrex.__path__)]
    modules.remove("pydrex.distributed")
    try:
        from pydrex import mesh
    except ImportError:
        modules.remove("pydrex.mesh")
    return modules


@pytest.mark.parametrize("module", _get_submodule_list())
def test_doctests(module, capsys):
    """Run doctests for all submodules."""
    with capsys.disabled():  # Pytest output capturing messes with doctests.
        _log.info("running doctests for %s...", module)
        try:
            n_fails, n_tests = doctest.testmod(
                importlib.import_module(module),
                raise_on_error=True,
                verbose=False,  # Change to True to debug doctest failures.
            )
            if n_fails > 0:
                raise AssertionError(f"there were {n_fails} doctest failures from {module}")
        except doctest.DocTestFailure as e:
            if e.test.lineno is None:
                lineno = ""
            else:
                lineno = f":{e.test.lineno + 1 + e.example.lineno}"
            raise AssertionError(
                f"{e.test.name} ({module}{lineno}) failed with:"
                + os.linesep
                + os.linesep
                + e.got
            ) from None
        except doctest.UnexpectedException as e:
            if e.test.lineno is None:
                lineno = ""
            else:
                lineno = f":{e.test.lineno + 1 + e.example.lineno}"
            err_type, err, _ = e.exc_info
            raise Error(
                f"{err_type.__qualname__} encountered in {e.test.name} ({module}{lineno})"
            ) from err



"""> PyDRex: Shims and wrappers for distributed-memory multiprocessing."""
import ray

from pydrex import diagnostics as _diagnostics

@ray.remote
def misorientation_indices(*args, **kwargs):
    return _diagnostics.misorientation_indices(*args, **kwargs)

@ray.remote
def misorientation_index(*args, **kwargs):
    return _diagnostics.misorientation_index(*args, **kwargs)



"""> PyDRex: Steady-state solutions of velocity (gradients) for various flows.

For the sake of consistency, all callables returned from methods in this module expect a
time scalar and 3D position vector as inputs. They also return 3D tensors in all cases.
This means they can be directly used as arguments to e.g.
`pydrex.minerals.Mineral.update_orientations`.

"""

import functools as ft

import numba as nb
import numpy as np

from pydrex import geometry as _geo


@nb.njit(fastmath=True)
def _simple_shear_2d_grad(t, x, direction, deformation_plane, strain_rate):
    grad_v = np.zeros((3, 3))
    grad_v[direction, deformation_plane] = 2 * strain_rate
    return grad_v


@nb.njit(fastmath=True)
def _simple_shear_2d(t, x, direction, deformation_plane, strain_rate):
    v = np.zeros(3)
    v[direction] = x[deformation_plane] * strain_rate
    return v


@nb.njit(fastmath=True)
def _cell_2d_grad(t, x, horizontal, vertical, velocity_edge, edge_length):
    _lim = edge_length / 2
    if np.abs(x[horizontal]) > _lim or np.abs(x[vertical]) > _lim:
        # NOTE: At the moment, this prints type info rather than values. Probably a
        # numba bug, version 0.59 shouldn't require these to be compile-time constants.
        raise ValueError(
            f"position ({x[horizontal]}, {x[vertical]}) is outside domain with xᵢ ∈ [-{_lim}, {_lim}]"
        )
    cos_πx_divd = np.cos(np.pi * x[horizontal] / edge_length)
    cos_πz_divd = np.cos(np.pi * x[vertical] / edge_length)
    sin_πx_divd = np.sin(np.pi * x[horizontal] / edge_length)
    sin_πz_divd = np.sin(np.pi * x[vertical] / edge_length)
    grad_v = np.zeros((3, 3))
    grad_v[horizontal, horizontal] = -np.pi / edge_length * sin_πz_divd * sin_πx_divd
    grad_v[horizontal, vertical] = np.pi / edge_length * cos_πz_divd * cos_πx_divd
    grad_v[vertical, vertical] = -np.pi / edge_length * cos_πz_divd * cos_πx_divd
    grad_v[vertical, horizontal] = np.pi / edge_length * sin_πz_divd * sin_πx_divd
    return velocity_edge * grad_v


@nb.njit(fastmath=True)
def _cell_2d(t, x, horizontal, vertical, velocity_edge, edge_length):
    _lim = edge_length / 2
    if np.abs(x[horizontal]) > _lim or np.abs(x[vertical]) > _lim:
        # NOTE: At the moment, this prints type info rather than values. Probably a
        # numba bug, version 0.59 shouldn't require these to be compile-time constants.
        raise ValueError(
            f"position ({x[horizontal]}, {x[vertical]}) is outside domain with xᵢ ∈ [-{_lim}, {_lim}]"
        )
    cos_πx_divd = np.cos(np.pi * x[horizontal] / edge_length)
    cos_πz_divd = np.cos(np.pi * x[vertical] / edge_length)
    sin_πx_divd = np.sin(np.pi * x[horizontal] / edge_length)
    sin_πz_divd = np.sin(np.pi * x[vertical] / edge_length)
    v = np.zeros(3)
    v[horizontal] = velocity_edge * cos_πx_divd * sin_πz_divd
    v[vertical] = -velocity_edge * sin_πx_divd * cos_πz_divd
    return v


@nb.njit(fastmath=True)
def _corner_2d(t, x, horizontal, vertical, plate_speed):
    h = x[horizontal]
    v = x[vertical]
    if np.abs(h) < 1e-15 and np.abs(v) < 1e-15:
        return np.full(3, np.nan)
    out = np.zeros(3)
    prefactor = 2 * plate_speed / np.pi
    out[horizontal] = prefactor * (np.arctan2(h, -v) + h * v / (h**2 + v**2))
    out[vertical] = prefactor * v**2 / (h**2 + v**2)
    return out


@nb.njit(fastmath=True)
def _corner_2d_grad(t, x, horizontal, vertical, plate_speed):
    h = x[horizontal]
    v = x[vertical]
    if np.abs(h) < 1e-15 and np.abs(v) < 1e-15:
        return np.full((3, 3), np.nan)
    grad_v = np.zeros((3, 3))
    prefactor = 4 * plate_speed / (np.pi * (h**2 + v**2) ** 2)
    grad_v[horizontal, horizontal] = -(h**2) * v
    grad_v[horizontal, vertical] = h**3
    grad_v[vertical, horizontal] = -h * v**2
    grad_v[vertical, vertical] = h**2 * v
    return prefactor * grad_v


def simple_shear_2d(direction, deformation_plane, strain_rate):
    """Return simple shear velocity and velocity gradient callables.

    The returned callables have signature f(t, x) where x is a 3D position vector.

    Args:
    - `direction` (one of {"X", "Y", "Z"}) — velocity vector direction
    - `deformation_plane` (one of {"X", "Y", "Z"}) — direction of velocity gradient
    - `strain_rate` (float) — 1/2 × magnitude of the largest eigenvalue of the velocity
      gradient

    .. note::
        Input arrays to the returned callables must have homogeneous element types.
        Arrays with e.g. both floating point and integer values are not supported.

    Examples:

    >>> import numpy as np
    >>> u, L = simple_shear_2d("X", "Z", 1e-4)
    >>> u(np.nan, np.array([0, 0, 0]))  # Time is not used here.
    array([0., 0., 0.])
    >>> u(np.nan, np.array([0, 0, 1]))
    array([0.0001, 0.    , 0.    ])
    >>> u(np.nan, np.array([0.0, 0.0, 2.0]))
    array([0.0002, 0.    , 0.    ])
    >>> L(np.nan, np.array([0, 0, 0]))
    array([[0.    , 0.    , 0.0002],
           [0.    , 0.    , 0.    ],
           [0.    , 0.    , 0.    ]])
    >>> L(np.nan, np.array([0.0, 0.0, 1.0]))
    array([[0.    , 0.    , 0.0002],
           [0.    , 0.    , 0.    ],
           [0.    , 0.    , 0.    ]])

    """
    try:
        indices = _geo.to_indices(direction, deformation_plane)
    except ValueError:
        raise ValueError(
            "unsupported shear type with"
            + f" direction = {direction}, deformation_plane = {deformation_plane}"
        )

    return (
        ft.partial(
            _simple_shear_2d,
            direction=indices[0],
            deformation_plane=indices[1],
            strain_rate=strain_rate,
        ),
        ft.partial(
            _simple_shear_2d_grad,
            direction=indices[0],
            deformation_plane=indices[1],
            strain_rate=strain_rate,
        ),
    )


def cell_2d(horizontal, vertical, velocity_edge, edge_length=2):
    r"""Get velocity and velocity gradient callables for a steady-state 2D Stokes cell.

    The cell is centered at (0,0) and the velocity field is defined by:
    $$
    \bm{u} = U\cos(π x/d)\sin(π z/d) \bm{\hat{h}} - U\sin(π x/d)\cos(π z/d) \bm{\hat{v}}
    $$
    where $\bm{\hat{h}}$ and $\bm{\hat{v}}$ are unit vectors in the chosen horizontal
    and vertical directions, respectively. The velocity at the cell edge has a magnitude
    of $U$ and $d$ is the length of a cell edge.

    The returned callables have signature f(t, x) where x is a 3D position vector.

    Args:
    - `horizontal` (one of {"X", "Y", "Z"}) — horizontal direction
    - `vertical` (one of {"X", "Y", "Z"}) — vertical direction
    - `velocity_edge` (float) — velocity magnitude at the center of the cell edge
    - `edge_length` (float, optional) — the edge length of the cell (= 2 by default)

    Examples:

    >>> import numpy as np
    >>> u, L = cell_2d("X", "Z", 1)
    >>> u(np.nan, np.array([0, 0, 0]))  # Time value is not used for steady flows.
    array([ 0.,  0., -0.])
    >>> u(np.nan, np.array([0, 0, 1]))
    array([ 1.,  0., -0.])
    >>> u(np.nan, np.array([0, 1, 0]))  # Y-value is not used.
    array([ 0.,  0., -0.])
    >>> u(np.nan, np.array([0, 0, -1]))
    array([-1.,  0., -0.])
    >>> u(np.nan, np.array([1, 0, 0]))
    array([ 0.,  0., -1.])
    >>> u(np.nan, np.array([-0.5, 0.0, 0.0]))
    array([0.        , 0.        , 0.70710678])
    >>> L(np.nan, np.array([0, 0, 0]))
    array([[-0.        ,  0.        ,  1.57079633],
           [ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        , -1.57079633]])
    >>> L(np.nan, np.array([0.5, 0.0, 0.0]))
    array([[-0.        ,  0.        ,  1.11072073],
           [ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        , -1.11072073]])
    >>> L(np.nan, np.array([0, 0, 0])) == L(np.nan, np.array([0, 1, 0]))  # Y-value is not used.
    array([[ True,  True,  True],
           [ True,  True,  True],
           [ True,  True,  True]])
    >>> L(np.nan, np.array([1, 0, 0])) == L(np.nan, np.array([0, 0, 1]))
    array([[ True,  True,  True],
           [ True,  True,  True],
           [ True,  True,  True]])
    >>> L(np.nan, np.array([1, 0, 0])) == L(np.nan, np.array([-1, 0, 0]))
    array([[ True,  True,  True],
           [ True,  True,  True],
           [ True,  True,  True]])
    >>> L(np.nan, np.array([1, 0, 0])) == L(np.nan, np.array([0, 0, -1]))
    array([[ True,  True,  True],
           [ True,  True,  True],
           [ True,  True,  True]])
    >>> L(np.nan, np.array([0.5, 0.0, 0.5]))
    array([[-0.78539816,  0.        ,  0.78539816],
           [ 0.        ,  0.        ,  0.        ],
           [ 0.78539816,  0.        , -0.78539816]])

    >>> u, L = cell_2d("X", "Z", 6.3e-10, 1e5)
    >>> u(np.nan, np.array([0, 0, 0]))
    array([ 0.,  0., -0.])
    >>> u(np.nan, np.array([0.0, 0.0, -5e4]))
    array([-6.3e-10,  0.0e+00, -0.0e+00])
    >>> u(np.nan, np.array([2e2, 0e0, 0e0]))
    array([ 0.0000000e+00,  0.0000000e+00, -3.9583807e-12])

    """
    if edge_length < 0:
        raise ValueError(f"edge length of 2D cell must be positive, not {edge_length}")
    try:
        indices = _geo.to_indices(horizontal, vertical)
    except ValueError:
        raise ValueError(
            "unsupported convection cell geometry with"
            + f" horizontal = {horizontal}, vertical = {vertical}"
        )

    return (
        ft.partial(
            _cell_2d,
            horizontal=indices[0],
            vertical=indices[1],
            velocity_edge=velocity_edge,
            edge_length=edge_length,
        ),
        ft.partial(
            _cell_2d_grad,
            horizontal=indices[0],
            vertical=indices[1],
            velocity_edge=velocity_edge,
            edge_length=edge_length,
        ),
    )


def corner_2d(horizontal, vertical, plate_speed):
    r"""Get velocity and velocity gradient callables for a steady-state 2D corner flow.

    The velocity field is defined by:
    $$
    \bm{u} = \frac{dr}{dt} \bm{\hat{r}} + r \frac{dθ}{dt} \bm{\hat{θ}}
    = \frac{2 U}{π}(θ\sinθ - \cosθ) ⋅ \bm{\hat{r}} + \frac{2 U}{π}θ\cosθ ⋅ \bm{\hat{θ}}
    $$
    where $θ = 0$ points vertically downwards along the ridge axis
    and $θ = π/2$ points along the surface. $U$ is the half spreading velocity.
    Streamlines for the flow obey:
    $$
    ψ = \frac{2 U r}{π}θ\cosθ
    $$
    and are related to the velocity through:
    $$
    \bm{u} = -\frac{1}{r} ⋅ \frac{dψ}{dθ} ⋅ \bm{\hat{r}} + \frac{dψ}{dr}\bm{\hat{θ}}
    $$
    Conversion to Cartesian ($x,y,z$) coordinates yields:
    $$
    \bm{u} = \frac{2U}{π} \left[
    \tan^{-1}\left(\frac{x}{-z}\right) + \frac{xz}{x^{2} + z^{2}} \right] \bm{\hat{x}} +
    \frac{2U}{π} \frac{z^{2}}{x^{2} + z^{2}} \bm{\hat{z}}
    $$
    where
    \begin{align\*}
    x &= r \sinθ \cr
    z &= -r \cosθ
    \end{align\*}
    and the velocity gradient is:
    $$
    L = \frac{4 U}{π{(x^{2}+z^{2})}^{2}} ⋅
    \begin{bmatrix}
        -x^{2}z & 0 & x^{3} \cr
        0 & 0 & 0 \cr
        -xz^{2} & 0 & x^{2}z
    \end{bmatrix}
    $$
    See also Fig. 5 in [Kaminski & Ribe, 2002](https://doi.org/10.1029/2001GC000222).

    The returned callables have signature f(t, x) where x is a 3D position vector.

    Args:
    - `horizontal` (one of {"X", "Y", "Z"}) — horizontal direction
    - `vertical` (one of {"X", "Y", "Z"}) — vertical direction
    - `plate_speed` (float) — speed of the “plate” i.e. upper boundary

    """
    try:
        indices = _geo.to_indices(horizontal, vertical)
    except ValueError:
        raise ValueError(
            "unsupported convection cell geometry with"
            + f" horizontal = {horizontal}, vertical = {vertical}"
        )

    return (
        ft.partial(
            _corner_2d,
            horizontal=indices[0],
            vertical=indices[1],
            plate_speed=plate_speed,
        ),
        ft.partial(
            _corner_2d_grad,
            horizontal=indices[0],
            vertical=indices[1],
            plate_speed=plate_speed,
        ),
    )



"""> PyDRex: Custom exceptions (subclasses of `pydrex.Error`)."""

# <https://docs.python.org/3.11/tutorial/errors.html#user-defined-exceptions>


class Error(Exception):
    """Base class for exceptions in PyDRex."""


class MissingDependencyError(Error):
    """Exception raised when optional dependencies are missing.

    Attributes:
        message — explanation of the error

    """

    def __init__(self, message):  # pylint: disable=super-init-not-called
        self.message = message


class ConfigError(Error):
    """Exception raised for errors in the input configuration.

    Attributes:
        message — explanation of the error

    """

    def __init__(self, message):  # pylint: disable=super-init-not-called
        self.message = message


class MeshError(Error):
    """Exception raised for errors in the input mesh.

    Attributes:
        message — explanation of the error

    """

    def __init__(self, message):  # pylint: disable=super-init-not-called
        self.message = message


class IterationError(Error):
    """Exception raised for errors in numerical iteration schemes.

    Attributes:
        message — explanation of the error

    """

    def __init__(self, message):  # pylint: disable=super-init-not-called
        # TODO: Add data attribute? Timestep?
        self.message = message


class SCSVError(Error):
    """Exception raised for errors in SCSV file I/O.

    Attributes:
    - message — explanation of the error

    """

    def __init__(self, message):  # pylint: disable=super-init-not-called
        self.message = message


class ModelContextError(Error):
    """Exception raised for errors in `mesh.Model` context state.

    Attributes:
        message — explanation of the error

    """

    def __init__(self, message):  # pylint: disable=super-init-not-called
        self.message = message



"""> PyDRex: tests for texture diagnostics."""

import numpy as np
import pytest
from numpy import random as rn
from scipy.spatial.transform import Rotation

from pydrex import diagnostics as _diagnostics
from pydrex import geometry as _geo
from pydrex import stats as _stats


class TestElasticityComponents:
    """Test symmetry decomposition of elastic tensors."""

    def test_olivine_Browaeys2004(self):
        C = np.array(
            [
                [192, 66, 60, 0, 0, 0],
                [66, 160, 56, 0, 0, 0],
                [60, 56, 272, 0, 0, 0],
                [0, 0, 0, 60, 0, 0],
                [0, 0, 0, 0, 62, 0],
                [0, 0, 0, 0, 0, 49],
            ]
        )
        out = _diagnostics.elasticity_components([C])
        # FIXME: How do they get 15.2% for hexagonal? It isn't the ratio of the norms
        # nor the ratio of the squared norms...
        expected = {
            "bulk_modulus": 109.8,
            "shear_modulus": 63.7,
            "percent_anisotropy": 20.7,
            # "percent_hexagonal": 15.2,
            # "percent_tetragonal": 0.625,
            # "percent_orthorhombic": 4.875,
            "percent_monoclinic": 0,
            "percent_triclinic": 0,
        }
        for k, v in expected.items():
            assert np.isclose(out[k][0], v, atol=0.1, rtol=0), f"{k}: {out[k]} != {v}"
        np.testing.assert_allclose(out["hexagonal_axis"][0], [0, 0, 1])

    def test_enstatite_Browaeys2004(self):
        C = np.array(
            [
                [225, 54, 72, 0, 0, 0],
                [54, 214, 53, 0, 0, 0],
                [72, 53, 178, 0, 0, 0],
                [0, 0, 0, 78, 0, 0],
                [0, 0, 0, 0, 82, 0],
                [0, 0, 0, 0, 0, 76],
            ]
        )
        out = _diagnostics.elasticity_components([C])
        # FIXME: Test remaining percentages when I figure out how they get the values.
        expected = {
            "bulk_modulus": 108.3,
            "shear_modulus": 76.4,
            "percent_anisotropy": 9.2,
            # "percent_hexagonal": 4.3,
            # "percent_tetragonal": ?,  # + ortho = 4.9
            # "percent_orthorhombic": ?,  # + tetra = 4.9
            "percent_monoclinic": 0,
            "percent_triclinic": 0,
        }
        for k, v in expected.items():
            assert np.isclose(out[k][0], v, atol=0.1, rtol=0), f"{k}: {out[k]} != {v}"
        np.testing.assert_allclose(out["hexagonal_axis"][0], [0, 0, 1])


class TestSymmetryPGR:
    """Test Point-Girdle-Random (eigenvalue) symmetry diagnostics."""

    def test_pointX(self):
        """Test diagnostics of point symmetry aligned to the X axis."""
        # Initial orientations within 10°.
        orientations = (
            Rotation.from_rotvec(
                np.stack(
                    [
                        [0, x * np.pi / 18 - np.pi / 36, x * np.pi / 18 - np.pi / 36]
                        for x in rn.default_rng().random(100)
                    ]
                )
            )
            .inv()
            .as_matrix()
        )
        np.testing.assert_allclose(
            _diagnostics.symmetry_pgr(orientations, axis="a"), (1, 0, 0), atol=0.05
        )

    def test_random(self):
        """Test diagnostics of random grain orientations."""
        orientations = Rotation.random(1000).as_matrix()
        np.testing.assert_allclose(
            _diagnostics.symmetry_pgr(orientations, axis="a"), (0, 0, 1), atol=0.15
        )

    def test_girdle(self):
        """Test diagnostics of girdled orientations."""
        rng = rn.default_rng()
        a = np.zeros(1000)
        b = np.zeros(1000)
        c = rng.normal(0, 1.0, size=1000)
        d = rng.normal(0, 1.0, size=1000)
        orientations = Rotation.from_quat(np.column_stack([a, b, c, d])).as_matrix()
        np.testing.assert_allclose(
            _diagnostics.symmetry_pgr(orientations, axis="a"), (0, 1, 0), atol=0.1
        )


class TestVolumeWeighting:
    """Tests for volumetric resampling of orientation data."""

    def test_output_shape(self):
        """Test that we get the correct output shape."""
        orientations = [
            Rotation.random(1000).as_matrix(),
            Rotation.random(1000).as_matrix(),
        ]
        fractions = [np.full(1000, 1 / 1000), np.full(1000, 1 / 1000)]
        new_orientations, new_fractions = _stats.resample_orientations(
            orientations, fractions
        )
        np.testing.assert_array_equal(
            np.asarray(orientations).shape, new_orientations.shape
        )
        np.testing.assert_array_equal(np.asarray(fractions).shape, new_fractions.shape)
        new_orientations, new_fractions = _stats.resample_orientations(
            orientations, fractions, n_samples=500
        )
        assert new_orientations.shape[1] == 500
        assert new_fractions.shape[1] == 500

    def test_upsample(self):
        """Test upsampling of the raw orientation data."""
        orientations = (
            Rotation.from_rotvec(
                [
                    [0, 0, 0],
                    [0, 0, np.pi / 6],
                    [np.pi / 6, 0, 0],
                ]
            )
            .inv()
            .as_matrix()
        )
        fractions = np.array([0.25, 0.6, 0.15])
        new_orientations, _ = _stats.resample_orientations(
            [orientations],
            [fractions],
            25,
        )
        assert np.all(a in orientations for a in new_orientations[0])

    def test_downsample(self):
        """Test downsampling of orientation data."""
        orientations = (
            Rotation.from_rotvec(
                [
                    [0, 0, 0],
                    [0, 0, np.pi / 6],
                    [np.pi / 6, 0, 0],
                ]
            )
            .inv()
            .as_matrix()
        )
        fractions = np.array([0.25, 0.6, 0.15])
        new_orientations = _stats.resample_orientations(
            [orientations],
            [fractions],
            2,
        )
        assert np.all(a in orientations for a in new_orientations[0])

    def test_common_input_errors(self):
        """Test that exceptions are raised for bad input data."""
        orientations = Rotation.from_rotvec(
            [
                [0, 0, 0],
                [0, 0, np.pi / 6],
                [np.pi / 6, 0, 0],
            ]
        )
        fractions = np.array([0.25, 0.6, 0.15])
        with pytest.raises(ValueError):
            # SciPy Rotation instance is not valid input.
            _stats.resample_orientations(orientations, fractions)
            # Input must be a stack of orientations.
            _stats.resample_orientations(orientations.as_matrix(), fractions)
            # Input must be a stack of fractions.
            _stats.resample_orientations([orientations.as_matrix()], fractions)
            # First two dimensions of inputs must match.
            _stats.resample_orientations([orientations.as_matrix()], [fractions[:-1]])

        # This is the proper way to do it:
        _stats.resample_orientations([orientations.as_matrix()], [fractions])


class TestBinghamStats:
    """Tests for antipodally symmetric (bingham) statistics."""

    def test_average_0(self):
        """Test bingham average of vectors aligned to the reference frame."""
        orientations = Rotation.from_rotvec([[0, 0, 0]] * 10).inv().as_matrix()
        a_mean = _diagnostics.bingham_average(orientations, axis="a")
        b_mean = _diagnostics.bingham_average(orientations, axis="b")
        c_mean = _diagnostics.bingham_average(orientations, axis="c")
        np.testing.assert_array_equal(a_mean, [1, 0, 0])
        np.testing.assert_array_equal(b_mean, [0, 1, 0])
        np.testing.assert_array_equal(c_mean, [0, 0, 1])

    def test_average_twopoles90Z(self):
        """Test bingham average of vectors rotated by ±90° around Z."""
        orientations = (
            Rotation.from_rotvec(
                [
                    [0, 0, -np.pi / 2],
                    [0, 0, np.pi / 2],
                ]
            )
            .inv()
            .as_matrix()
        )
        a_mean = _diagnostics.bingham_average(orientations, axis="a")
        b_mean = _diagnostics.bingham_average(orientations, axis="b")
        c_mean = _diagnostics.bingham_average(orientations, axis="c")
        np.testing.assert_array_equal(a_mean, [0, 1, 0])
        np.testing.assert_array_equal(b_mean, [1, 0, 0])
        np.testing.assert_array_equal(c_mean, [0, 0, 1])

    def test_average_spread10X(self):
        """Test bingham average of vectors spread within 10° of the ±X-axis."""
        orientations = (
            Rotation.from_rotvec(
                np.stack(
                    [
                        [0, x * np.pi / 18 - np.pi / 36, x * np.pi / 18 - np.pi / 36]
                        for x in rn.default_rng().random(100)
                    ]
                )
            )
            .inv()
            .as_matrix()
        )
        a_mean = _diagnostics.bingham_average(orientations, axis="a")
        b_mean = _diagnostics.bingham_average(orientations, axis="b")
        c_mean = _diagnostics.bingham_average(orientations, axis="c")
        np.testing.assert_allclose(np.abs(a_mean), [1.0, 0, 0], atol=np.sin(np.pi / 18))
        np.testing.assert_allclose(np.abs(b_mean), [0, 1.0, 0], atol=np.sin(np.pi / 18))
        np.testing.assert_allclose(np.abs(c_mean), [0, 0, 1.0], atol=np.sin(np.pi / 18))


class TestMIndex:
    """Tests for the M-index texture strength diagnostic."""

    # TODO: Add tests for other (non-orthorhombic) lattice symmetries.

    def test_texture_uniform_ortho(self, seed):
        """Test with random (uniform distribution) orthorhombic grain orientations."""
        orientations = Rotation.random(1000, random_state=seed).as_matrix()
        assert np.isclose(
            _diagnostics.misorientation_index(
                orientations, _geo.LatticeSystem.orthorhombic
            ),
            0.05,
            atol=1e-2,
            rtol=0,
        )

    def test_texture_spread10X_ortho(self, seed):
        """Test for orthorhombic grains spread within 10° of the ±X axis."""
        orientations = (
            Rotation.from_rotvec(
                np.stack(
                    [
                        [0, x * np.pi / 18 - np.pi / 36, x * np.pi / 18 - np.pi / 36]
                        for x in rn.default_rng(seed=seed).random(1000)
                    ]
                )
            )
            .inv()
            .as_matrix()
        )
        assert np.isclose(
            _diagnostics.misorientation_index(
                orientations, _geo.LatticeSystem.orthorhombic
            ),
            0.99,
            atol=1e-2,
            rtol=0,
        )

    def test_texture_spread45X_ortho(self, seed):
        """Test for orthorhombic grains spread within 45° of the ±X axis."""
        orientations = Rotation.from_rotvec(
            np.stack(
                [
                    [0, x * np.pi / 2 - np.pi / 4, x * np.pi / 2 - np.pi / 4]
                    for x in rn.default_rng(seed=seed).random(1000)
                ]
            )
        ).as_matrix()
        assert np.isclose(
            _diagnostics.misorientation_index(
                orientations, _geo.LatticeSystem.orthorhombic
            ),
            0.81,
            atol=1e-2,
            rtol=0,
        )

    def test_textures_increasing_ortho(self, seed):
        """Test M-index for textures of increasing strength."""
        M_vals = np.empty(4)
        factors = (2, 3, 4, 5)
        for i, factor in enumerate(factors):
            orientations = Rotation.from_rotvec(
                np.stack(
                    [
                        [
                            0,
                            x * np.pi / factor - np.pi / factor / 2,
                            x * np.pi / factor - np.pi / factor / 2,
                        ]
                        for x in rn.default_rng(seed=seed).random(1000)
                    ]
                )
            ).as_matrix()
            M_vals[i] = _diagnostics.misorientation_index(
                orientations, _geo.LatticeSystem.orthorhombic
            )
        assert np.all(
            np.diff(M_vals) >= 0
        ), f"M-index values {M_vals} are not increasing"

    def test_texture_girdle_ortho(self, seed):
        """Test M-index for girdled texture."""
        rng = rn.default_rng(seed=seed)
        a = np.zeros(1000)
        b = np.zeros(1000)
        c = rng.normal(0, 1.0, size=1000)
        d = rng.normal(0, 1.0, size=1000)
        orientations = Rotation.from_quat(np.column_stack([a, b, c, d])).as_matrix()
        assert np.isclose(
            _diagnostics.misorientation_index(
                orientations, _geo.LatticeSystem.orthorhombic
            ),
            0.67,
            atol=1e-2,
        )
